import os
import time
import datetime
import argparse
import yaml

import numpy as np
from scipy import sparse
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import DataLoader
import torch.optim as optim
from torch.optim.lr_scheduler import ReduceLROnPlateau
import torch.backends.cudnn as cudnn

from graph4nlp.pytorch.datasets.squad import SQuADDataset
from graph4nlp.pytorch.data.data import from_batch
from graph4nlp.pytorch.modules.graph_construction import *
from graph4nlp.pytorch.modules.graph_construction.embedding_construction import RNNEmbedding, WordEmbedding
from graph4nlp.pytorch.modules.graph_embedding import GAT, GraphSAGE, GGNN
from graph4nlp.pytorch.modules.prediction.generation.StdRNNDecoder import StdRNNDecoder
from graph4nlp.pytorch.modules.utils.generic_utils import grid, to_cuda, dropout_fn, sparse_mx_to_torch_sparse_tensor, EarlyStopping
from examples.pytorch.semantic_parsing.graph2seq.nouse_loss import Graph2SeqLoss, CoverageLoss
from graph4nlp.pytorch.modules.evaluation import BLEU, METEOR, ROUGE
from graph4nlp.pytorch.modules.utils.logger import Logger
from graph4nlp.pytorch.modules.utils import constants as Constants
import torch.multiprocessing
torch.multiprocessing.set_sharing_strategy('file_system')


class QGModel(nn.Module):
    def __init__(self, vocab, config):
        super(QGModel, self).__init__()
        self.config = config
        self.vocab = vocab
        embedding_style = {'single_token_item': True if config['graph_type'] != 'ie' else False,
                            'emb_strategy': config.get('emb_strategy', 'w2v_bilstm'),
                            'num_rnn_layers': 1,
                            'bert_model_name': config.get('bert_model_name', 'bert-base-uncased'),
                            'bert_lower_case': True
                           }
        print('embedding_style: {}'.format(embedding_style))

        assert not (config['graph_type'] in ('node_emb', 'node_emb_refined') and config['gnn'] == 'gat'), \
                                'dynamic graph construction does not support GAT'

        use_edge_weight = False
        if config['graph_type'] == 'dependency':
            self.graph_topology = DependencyBasedGraphConstruction(embedding_style=embedding_style,
                                                                   vocab=vocab.in_word_vocab,
                                                                   hidden_size=config['num_hidden'],
                                                                   word_dropout=config['word_dropout'],
                                                                   rnn_dropout=config['enc_rnn_dropout'],
                                                                   fix_word_emb=not config['no_fix_word_emb'],
                                                                   fix_bert_emb=not config.get('no_fix_bert_emb', False),
                                                                   device=config['device'])
        elif config['graph_type'] == 'constituency':
            self.graph_topology = ConstituencyBasedGraphConstruction(embedding_style=embedding_style,
                                                                   vocab=vocab.in_word_vocab,
                                                                   hidden_size=config['num_hidden'],
                                                                   word_dropout=config['word_dropout'],
                                                                   rnn_dropout=config['enc_rnn_dropout'],
                                                                   fix_word_emb=not config['no_fix_word_emb'],
                                                                   device=config['device'])
        elif config['graph_type'] == 'ie':
            self.graph_topology = IEBasedGraphConstruction(embedding_style=embedding_style,
                                                                   vocab=vocab.in_word_vocab,
                                                                   hidden_size=config['num_hidden'],
                                                                   word_dropout=config['word_dropout'],
                                                                   rnn_dropout=config['enc_rnn_dropout'],
                                                                   fix_word_emb=not config['no_fix_word_emb'],
                                                                   device=config['device'])
        elif config['graph_type'] == 'node_emb':
            self.graph_topology = NodeEmbeddingBasedGraphConstruction(
                                    vocab.in_word_vocab,
                                    embedding_style,
                                    sim_metric_type=config['gl_metric_type'],
                                    num_heads=config['gl_num_heads'],
                                    top_k_neigh=config['gl_top_k'],
                                    epsilon_neigh=config['gl_epsilon'],
                                    smoothness_ratio=config['gl_smoothness_ratio'],
                                    connectivity_ratio=config['gl_connectivity_ratio'],
                                    sparsity_ratio=config['gl_sparsity_ratio'],
                                    input_size=config['num_hidden'],
                                    hidden_size=config['gl_num_hidden'],
                                    fix_word_emb=not config['no_fix_word_emb'],
                                    word_dropout=config['word_dropout'],
                                    rnn_dropout=config['enc_rnn_dropout'],
                                    device=config['device'])
            use_edge_weight = True
        elif config['graph_type'] == 'node_emb_refined':
            self.graph_topology = NodeEmbeddingBasedRefinedGraphConstruction(
                                    vocab.in_word_vocab,
                                    embedding_style,
                                    config['init_adj_alpha'],
                                    sim_metric_type=config['gl_metric_type'],
                                    num_heads=config['gl_num_heads'],
                                    top_k_neigh=config['gl_top_k'],
                                    epsilon_neigh=config['gl_epsilon'],
                                    smoothness_ratio=config['gl_smoothness_ratio'],
                                    connectivity_ratio=config['gl_connectivity_ratio'],
                                    sparsity_ratio=config['gl_sparsity_ratio'],
                                    input_size=config['num_hidden'],
                                    hidden_size=config['gl_num_hidden'],
                                    fix_word_emb=not config['no_fix_word_emb'],
                                    word_dropout=config['word_dropout'],
                                    rnn_dropout=config['enc_rnn_dropout'],
                                    device=config['device'])
            use_edge_weight = True
        else:
            raise RuntimeError('Unknown graph_type: {}'.format(config['graph_type']))


        if 'w2v' in self.graph_topology.embedding_layer.word_emb_layers:
            self.word_emb = self.graph_topology.embedding_layer.word_emb_layers['w2v'].word_emb_layer
        else:
            self.word_emb = WordEmbedding(
                            self.vocab.in_word_vocab.embeddings.shape[0],
                            self.vocab.in_word_vocab.embeddings.shape[1],
                            pretrained_word_emb=self.vocab.in_word_vocab.embeddings,
                            fix_emb=not config['no_fix_word_emb'],
                            device=config['device']).word_emb_layer

        answer_feat_size = self.vocab.in_word_vocab.embeddings.shape[1]
        if 'node_edge_bert' in self.graph_topology.embedding_layer.word_emb_layers:
            self.bert_encoder = self.graph_topology.embedding_layer.word_emb_layers['node_edge_bert']
            answer_feat_size += self.bert_encoder.bert_model.config.hidden_size
        elif 'seq_bert' in self.graph_topology.embedding_layer.word_emb_layers:
            self.bert_encoder = self.graph_topology.embedding_layer.word_emb_layers['seq_bert']
            answer_feat_size += self.bert_encoder.bert_model.config.hidden_size
        else:
            self.bert_encoder = None


        self.ans_rnn_encoder = RNNEmbedding(
                                    answer_feat_size,
                                    config['num_hidden'],
                                    bidirectional=True,
                                    num_layers=1,
                                    rnn_type='lstm',
                                    dropout=config['enc_rnn_dropout'],
                                    device=config['device'])

        # soft-alignment between context and answer
        self.ctx2ans_attn = Context2AnswerAttention(config['num_hidden'], config['num_hidden'])
        self.fuse_ctx_ans = nn.Linear(2 * config['num_hidden'], config['num_hidden'], bias=False)


        if config['gnn'] == 'gat':
            heads = [config['gat_num_heads']] * (config['gnn_num_layers'] - 1) + [config['gat_num_out_heads']]
            self.gnn = GAT(config['gnn_num_layers'],
                        config['num_hidden'],
                        config['num_hidden'],
                        config['num_hidden'],
                        heads,
                        direction_option=config['gnn_direction_option'],
                        feat_drop=config['gnn_dropout'],
                        attn_drop=config['gat_attn_dropout'],
                        negative_slope=config['gat_negative_slope'],
                        residual=config['gat_residual'],
                        activation=F.elu)
        elif config['gnn'] == 'graphsage':
            self.gnn = GraphSAGE(config['gnn_num_layers'],
                        config['num_hidden'],
                        config['num_hidden'],
                        config['num_hidden'],
                        config['graphsage_aggreagte_type'],
                        direction_option=config['gnn_direction_option'],
                        feat_drop=config['gnn_dropout'],
                        bias=True,
                        norm=None,
                        activation=F.relu,
                        use_edge_weight=use_edge_weight)
        elif config['gnn'] == 'ggnn':
            self.gnn = GGNN(config['gnn_num_layers'],
                        config['num_hidden'],
                        config['num_hidden'],
                        config['num_hidden'],
                        feat_drop=config['gnn_dropout'],
                        direction_option=config['gnn_direction_option'],
                        bias=True,
                        use_edge_weight=use_edge_weight)
        else:
            raise RuntimeError('Unknown gnn type: {}'.format(config['gnn']))

        self.seq_decoder = StdRNNDecoder(config['max_dec_steps'],
                                         2 * config['num_hidden'] if config['gnn_direction_option'] == 'bi_sep' else config['num_hidden'],
                                         config['num_hidden'],
                                         self.word_emb,
                                         self.vocab.in_word_vocab,
                                         graph_pooling_strategy=config['graph_pooling_strategy'],
                                         attention_type=config['dec_attention_type'],
                                         rnn_emb_input_size=config['num_hidden'],
                                         node_type_num=None,
                                         fuse_strategy=config['dec_fuse_strategy'],
                                         use_copy=config.get('use_copy', False),
                                         use_coverage=config['use_coverage'],
                                         coverage_strategy=config.get('coverage_strategy', 'sum'),
                                         tgt_emb_as_output_layer=config['tgt_emb_as_output_layer'],
                                         dropout=config['dec_rnn_dropout'])
        self.loss_calc = Graph2SeqLoss(self.vocab.in_word_vocab)
        self.loss_cover = CoverageLoss(config['coverage_loss_ratio'])


    def forward(self, data, require_loss=True):
        num_graph_nodes = []
        for g in data['graph_data']:
            num_graph_nodes.append(g.get_node_num())

        # graph embedding construction
        batch_gd = self.graph_topology(data['graph_data'])

        # answer alignment
        answer_feat = self.word_emb(data['input_tensor2'])
        answer_feat = dropout_fn(answer_feat, self.config['word_dropout'], shared_axes=[-2], training=self.training)

        if self.bert_encoder is not None:
            answer_bert_feat = self.bert_encoder(data['input_text2'])
            answer_feat = torch.cat([answer_feat, answer_bert_feat], -1)

        answer_feat = self.ans_rnn_encoder(answer_feat, data['input_length2'])[0]
        new_node_feat = self.answer_alignment(batch_gd.node_features['node_feat'], answer_feat, answer_feat, num_graph_nodes, data['input_length2'])
        batch_gd.node_features['node_feat'] = new_node_feat

        # run GNN
        self.gnn(batch_gd)
        batch_gd.node_features['rnn_emb'] = batch_gd.node_features['node_feat']

        # seq decoder
        prob, enc_attn_weights, coverage_vectors = self.seq_decoder(from_batch(batch_gd), tgt_seq=data['tgt_tensor'])
        if require_loss:
            loss = self.loss_calc(prob, data['tgt_tensor'])
            loss += self.loss_cover(enc_attn_weights, coverage_vectors)
            return prob, loss
        else:
            return prob


    def answer_alignment(self, node_feat, answer_feat, out_answer_feat, num_nodes, answer_length):
        assert len(num_nodes) == len(answer_length)
        mask = []
        for i in range(len(num_nodes)): # batch
            tmp_mask = np.ones((num_nodes[i], answer_feat.shape[1]))
            tmp_mask[:, answer_length[i]:] = 0
            mask.append(sparse.coo_matrix(tmp_mask))

        mask = sparse.block_diag(mask)
        mask = to_cuda(sparse_mx_to_torch_sparse_tensor(mask).to_dense(), self.config['device'])

        answer_feat = answer_feat.reshape(-1, answer_feat.shape[-1])
        out_answer_feat = out_answer_feat.reshape(-1, out_answer_feat.shape[-1])
        ctx_aware_ans_feat = self.ctx2ans_attn(node_feat, answer_feat, out_answer_feat, mask)
        new_node_feat = self.fuse_ctx_ans(torch.cat([node_feat, ctx_aware_ans_feat], -1))

        return new_node_feat


class Context2AnswerAttention(nn.Module):
    def __init__(self, dim, hidden_size):
        super(Context2AnswerAttention, self).__init__()
        self.linear_sim = nn.Linear(dim, hidden_size, bias=False)

    def forward(self, context, answers, out_answers, mask=None):
        """
        Parameters
        :context, (L, dim)
        :answers, (N, dim)
        :out_answers, (N, dim)
        :mask, (L, N)

        Returns
        :ques_emb, (L, dim)
        """
        context_fc = torch.relu(self.linear_sim(context))
        questions_fc = torch.relu(self.linear_sim(answers))

        # shape: (L, N)
        attention = torch.matmul(context_fc, questions_fc.transpose(-1, -2))
        if mask is not None:
            attention = attention.masked_fill_(~mask.bool(), -Constants.INF)
        prob = torch.softmax(attention, dim=-1)
        # shape: (L, dim)
        emb = torch.matmul(prob, out_answers)

        return emb


class ModelHandler:
    def __init__(self, config):
        super(ModelHandler, self).__init__()
        self.config = config
        self.logger = Logger(config['out_dir'], config={k:v for k, v in config.items() if k != 'device'}, overwrite=True)
        self.logger.write(config['out_dir'])
        self._build_dataloader()
        self._build_model()
        self._build_optimizer()
        self._build_evaluation()

    def _build_dataloader(self):
        dynamic_init_topology_builder = None
        if self.config['graph_type'] == 'dependency':
            topology_builder = DependencyBasedGraphConstruction
            graph_type = 'static'
            merge_strategy = 'tailhead'
        elif self.config['graph_type'] == 'constituency':
            topology_builder = ConstituencyBasedGraphConstruction
            graph_type = 'static'
            merge_strategy = 'tailhead'
        elif self.config['graph_type'] == 'ie':
            topology_builder = IEBasedGraphConstruction
            graph_type = 'static'
            merge_strategy = 'global'
        elif self.config['graph_type'] == 'node_emb':
            topology_builder = NodeEmbeddingBasedGraphConstruction
            graph_type = 'dynamic'
            merge_strategy = None
        elif self.config['graph_type'] == 'node_emb_refined':
            topology_builder = NodeEmbeddingBasedRefinedGraphConstruction
            graph_type = 'dynamic'
            merge_strategy = 'tailhead'

            if self.config['init_graph_type'] == 'line':
                dynamic_init_topology_builder = None
            elif self.config['init_graph_type'] == 'dependency':
                dynamic_init_topology_builder = DependencyBasedGraphConstruction
            elif self.config['init_graph_type'] == 'constituency':
                dynamic_init_topology_builder = ConstituencyBasedGraphConstruction
            elif self.config['init_graph_type'] == 'ie':
                merge_strategy = 'global'
                dynamic_init_topology_builder = IEBasedGraphConstruction
            else:
                # dynamic_init_topology_builder
                raise RuntimeError('Define your own dynamic_init_topology_builder')
        else:
            raise RuntimeError('Unknown graph_type: {}'.format(self.config['graph_type']))

        topology_subdir = '{}_based_graph'.format(self.config['graph_type'])
        if self.config['graph_type'] == 'node_emb_refined':
            topology_subdir += '_{}'.format(self.config['init_graph_type'])

        dataset = SQuADDataset(root_dir=self.config.get('root_dir', 'examples/pytorch/question_generation/data/squad_split2'),
                              pretrained_word_emb_file=self.config['pre_word_emb_file'],
                              merge_strategy=merge_strategy,
                              seed=self.config['seed'],
                              graph_type=graph_type,
                              topology_builder=topology_builder,
                              topology_subdir=topology_subdir,
                              dynamic_graph_type=self.config['graph_type'] if self.config['graph_type'] in ('node_emb', 'node_emb_refined') else None,
                              dynamic_init_topology_builder=dynamic_init_topology_builder,
                              dynamic_init_topology_aux_args={'dummy_param': 0})

        # TODO: use small ratio of the data
        dataset.train = dataset.train[:self.config['n_samples']]
        dataset.val = dataset.val[:self.config['n_samples']]
        dataset.test = dataset.test[:self.config['n_samples']]

        self.train_dataloader = DataLoader(dataset.train, batch_size=self.config['batch_size'], shuffle=True,
                                           num_workers=self.config['num_workers'],
                                           collate_fn=dataset.collate_fn)
        self.val_dataloader = DataLoader(dataset.val, batch_size=self.config['batch_size'], shuffle=False,
                                          num_workers=self.config['num_workers'],
                                          collate_fn=dataset.collate_fn)
        self.test_dataloader = DataLoader(dataset.test, batch_size=self.config['batch_size'], shuffle=False,
                                          num_workers=self.config['num_workers'],
                                          collate_fn=dataset.collate_fn)
        self.vocab = dataset.vocab_model
        self.num_train = len(dataset.train)
        self.num_val = len(dataset.val)
        self.num_test = len(dataset.test)
        print('Train size: {}, Val size: {}, Test size: {}'
            .format(self.num_train, self.num_val, self.num_test))
        self.logger.write('Train size: {}, Val size: {}, Test size: {}'
            .format(self.num_train, self.num_val, self.num_test))

    def _build_model(self):
        self.model = QGModel(self.vocab, self.config).to(self.config['device'])

    def _build_optimizer(self):
        parameters = [p for p in self.model.parameters() if p.requires_grad]
        self.optimizer = optim.Adam(parameters, lr=self.config['lr'])
        self.stopper = EarlyStopping(os.path.join(self.config['out_dir'], Constants._SAVED_WEIGHTS_FILE), patience=self.config['patience'])
        self.scheduler = ReduceLROnPlateau(self.optimizer, mode='max', factor=self.config['lr_reduce_factor'], \
            patience=self.config['lr_patience'], verbose=True)

    def _build_evaluation(self):
        self.metrics = {'BLEU': BLEU(n_grams=[1, 2, 3, 4]),
                        'METEOR': METEOR(),
                        'ROUGE': ROUGE()}

    def train(self):
        dur = []
        for epoch in range(self.config['epochs']):
            self.model.train()
            train_loss = []
            t0 = time.time()
            for i, data in enumerate(self.train_dataloader):
                data = all_to_cuda(data, self.config['device'])
                logits, loss = self.model(data, require_loss=True)
                self.optimizer.zero_grad()
                loss.backward()
                if self.config.get('grad_clipping', None) not in (None, 0):
                    # Clip gradients
                    parameters = [p for p in self.model.parameters() if p.requires_grad]
                    # if self.config['use_bert'] and self.config.get('finetune_bert', None):
                    #     parameters += [p for p in self.config['bert_model'].parameters() if p.requires_grad]

                    torch.nn.utils.clip_grad_norm_(parameters, self.config['grad_clipping'])

                self.optimizer.step()
                train_loss.append(loss.item())

                pred = torch.max(logits, dim=-1)[1].cpu()
                dur.append(time.time() - t0)

            val_scores = self.evaluate(self.val_dataloader)
            self.scheduler.step(val_scores[self.config['early_stop_metric']])
            format_str = 'Epoch: [{} / {}] | Time: {:.2f}s | Loss: {:.4f} | Val scores:'.format(epoch + 1, self.config['epochs'], np.mean(dur), np.mean(train_loss))
            format_str += self.metric_to_str(val_scores)
            print(format_str)
            self.logger.write(format_str)

            if self.stopper.step(val_scores[self.config['early_stop_metric']], self.model):
                break

        return self.stopper.best_score

    def evaluate(self, dataloader):
        self.model.eval()
        with torch.no_grad():
            pred_collect = []
            gt_collect = []
            for i, data in enumerate(dataloader):
                data = all_to_cuda(data, self.config['device'])
                prob = self.model(data, require_loss=False)
                pred = prob.argmax(dim=-1)

                pred_str = wordid2str(pred.detach().cpu(), self.vocab.in_word_vocab)
                tgt_str = wordid2str(data['tgt_tensor'], self.vocab.in_word_vocab)
                pred_collect.extend(pred_str)
                gt_collect.extend(data['tgt_text'])

            scores = self.evaluate_predictions(gt_collect, pred_collect)

            return scores

    def test(self):
        # restored best saved model
        self.stopper.load_checkpoint(self.model)

        t0 = time.time()
        scores = self.evaluate(self.test_dataloader)
        dur = time.time() - t0
        format_str = 'Test examples: {} | Time: {:.2f}s |  Test scores:'.format(self.num_test, dur)
        format_str += self.metric_to_str(scores)
        print(format_str)
        self.logger.write(format_str)

        return scores

    def evaluate_predictions(self, ground_truth, predict):
        output = {}
        for name, scorer in self.metrics.items():
            score = scorer.calculate_scores(ground_truth=ground_truth, predict=predict)
            if name.upper() == 'BLEU':
                for i in range(len(score[0])):
                    output['BLEU_{}'.format(i + 1)] = score[0][i]
            else:
                output[name] = score[0]

        return output

    def metric_to_str(self, metrics):
        format_str = ''
        for k in metrics:
            format_str += ' {} = {:0.5f},'.format(k.upper(), metrics[k])

        return format_str[:-1]


def main(config):
    # configure
    np.random.seed(config['seed'])
    torch.manual_seed(config['seed'])

    if not config['no_cuda'] and torch.cuda.is_available():
        print('[ Using CUDA ]')
        config['device'] = torch.device('cuda' if config['gpu'] < 0 else 'cuda:%d' % config['gpu'])
        cudnn.benchmark = True
        torch.cuda.manual_seed(config['seed'])
    else:
        config['device'] = torch.device('cpu')

    ts = datetime.datetime.now().timestamp()
    config['out_dir'] += '_{}'.format(ts)
    print('\n' + config['out_dir'])

    runner = ModelHandler(config)
    t0 = time.time()

    val_score = runner.train()
    test_scores = runner.test()

    print('Removed best saved model file to save disk space')
    os.remove(runner.stopper.save_model_path)
    runtime = time.time() - t0
    print('Total runtime: {:.2f}s'.format(time.time() - t0))
    runner.logger.write('Total runtime: {:.2f}s\n'.format(runtime))
    runner.logger.close()

    return val_score, test_scores

def wordid2str(word_ids, vocab):
    ret = []
    assert len(word_ids.shape) == 2, print(word_ids.shape)
    for i in range(word_ids.shape[0]):
        id_list = word_ids[i, :]
        ret_inst = []
        for j in range(id_list.shape[0]):
            if id_list[j] == vocab.EOS:
                break
            token = vocab.getWord(id_list[j])
            ret_inst.append(token)
        ret.append(' '.join(ret_inst))

    return ret

def all_to_cuda(data, device=None):
    if isinstance(data, torch.Tensor):
        data = to_cuda(data, device)
    elif isinstance(data, (list, dict)):
        keys = range(len(data)) if isinstance(data, list) else data.keys()
        for k in keys:
            if isinstance(data[k], torch.Tensor):
                data[k] = to_cuda(data[k], device)

    return data

################################################################################
# ArgParse and Helper Functions #
################################################################################
def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-config', '--config', required=True, type=str, help='path to the config file')
    parser.add_argument('--grid_search', action='store_true', help='flag: grid search')
    args = vars(parser.parse_args())

    return args


def get_config(config_path='config.yml'):
    with open(config_path, 'r') as setting:
        config = yaml.load(setting)

    return config


def print_config(config):
    print('**************** MODEL CONFIGURATION ****************')
    for key in sorted(config.keys()):
        val = config[key]
        keystr = '{}'.format(key) + (' ' * (24 - len(key)))
        print('{} -->   {}'.format(keystr, val))
    print('**************** MODEL CONFIGURATION ****************')


def grid_search_main(config):
    grid_search_hyperparams = []
    log_path = config['out_dir']
    for k, v in config.items():
        if isinstance(v, list):
            grid_search_hyperparams.append(k)
            log_path += '_{}_{}'.format(k, v)

    logger = Logger(log_path, config=config, overwrite=True)

    best_config = None
    best_score = -1
    best_scores = None
    configs = grid(config)
    for cnf in configs:
        for k in grid_search_hyperparams:
            cnf['out_dir'] += '_{}_{}'.format(k, cnf[k])

        val_score, test_scores = main(cnf)
        if best_score < test_scores[cnf['early_stop_metric']]:
            best_score = test_scores[cnf['early_stop_metric']]
            best_scores = test_scores
            best_config = cnf
            print('Found a better configuration: {}'.format(best_scores))
            logger.write('Found a better configuration: {}'.format(best_scores))

    print('\nBest configuration:')
    logger.write('\nBest configuration:')
    for k in grid_search_hyperparams:
        print('{}: {}'.format(k, best_config[k]))
        logger.write('{}: {}'.format(k, best_config[k]))

    print('Best score: {}'.format(best_scores))
    logger.write('Best score: {}\n'.format(best_scores))
    logger.close()


if __name__ == '__main__':
    cfg = get_args()
    config = get_config(cfg['config'])
    print_config(config)
    if cfg['grid_search']:
        grid_search_main(config)
    else:
        main(config)



from stanfordcorenlp import StanfordCoreNLP


processor = StanfordCoreNLP('http://localhost', port=9000, timeout=1000)
props = {
            'annotators': 'ssplit,tokenize,depparse',
            "tokenize.options":
                "splitHyphenated=false,normalizeParentheses=true,normalizeOtherBrackets=true",
            "tokenize.whitespace": False,
            'ssplit.isOneSentence': False,
            'outputFormat': 'json'
        }
raw_text_data = "James went to the corner-shop."
dep_json = processor.annotate(raw_text_data.strip(), properties=props)
print(dep_json)


import dgl
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader

from graph4nlp.pytorch.data.data import from_batch
from graph4nlp.pytorch.datasets.jobs import JobsDataset
from graph4nlp.pytorch.modules.evaluation.base import EvaluationMetricBase
from graph4nlp.pytorch.modules.graph_construction.dependency_graph_construction import DependencyBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_embedding.gat import GAT
from graph4nlp.pytorch.modules.prediction.generation.StdRNNDecoder import StdRNNDecoder
from graph4nlp.pytorch.modules.utils.vocab_utils import Vocab


class ExpressionAccuracy(EvaluationMetricBase):
    def __init__(self):
        super(ExpressionAccuracy, self).__init__()

    def calculate_scores(self, ground_truth, predict):
        correct = 0
        assert len(ground_truth) == len(predict)
        for gt, pred in zip(ground_truth, predict):
            print("ground truth: ", gt)
            print("prediction: ", pred)

            if gt == pred:
                correct += 1.
        return correct / len(ground_truth)


def logits2seq(prob: torch.Tensor):
    ids = prob.argmax(dim=-1)
    return ids


def wordid2str(word_ids, vocab: Vocab):
    ret = []
    assert len(word_ids.shape) == 2, print(word_ids.shape)
    for i in range(word_ids.shape[0]):
        id_list = word_ids[i, :]
        ret_inst = []
        for j in range(id_list.shape[0]):
            if id_list[j] == vocab.EOS:
                break
            token = vocab.getWord(id_list[j])
            ret_inst.append(token)
        ret.append(" ".join(ret_inst))
    return ret


class Graph2seqLoss(nn.Module):
    def __init__(self, vocab: Vocab):
        super(Graph2seqLoss, self).__init__()
        self.loss_func = nn.NLLLoss()
        self.VERY_SMALL_NUMBER = 1e-31
        self.vocab = vocab

    def forward(self, prob, gt):
        assert prob.shape[0:1] == gt.shape[0:1]
        assert len(prob.shape) == 3
        log_prob = torch.log(prob + self.VERY_SMALL_NUMBER)
        batch_size = gt.shape[0]
        step = gt.shape[1]

        mask = 1 - gt.data.eq(self.vocab.PAD).float()

        prob_select = torch.gather(log_prob.view(batch_size * step, -1), 1, gt.view(-1, 1))

        prob_select_masked = - torch.masked_select(prob_select, mask.view(-1, 1).byte())
        loss = torch.mean(prob_select_masked)
        return loss


class Graph2seq(nn.Module):
    def __init__(self, vocab, hidden_size=300, direction_option='bi_sep', device='cpu'):
        super(Graph2seq, self).__init__()

        use_cuda = (device == 'cuda')
        self.vocab = vocab
        embedding_style = {'word_emb_type': 'w2v', 'node_edge_emb_strategy': "mean",
                           'seq_info_encode_strategy': "bilstm"}
        self.graph_topology = DependencyBasedGraphConstruction(embedding_style=embedding_style,
                                                               vocab=vocab.in_word_vocab,
                                                               hidden_size=hidden_size, dropout=0.2, use_cuda=use_cuda,
                                                               fix_word_emb=False)
        self.gnn = None
        self.word_emb = self.graph_topology.embedding_layer.word_emb_layers[0].word_emb_layer
        self.gnn_encoder = GAT(2, hidden_size, hidden_size, hidden_size, [2, 1], direction_option=direction_option).to(
            device)
        self.seq_decoder = StdRNNDecoder(max_decoder_step=50,
                                         decoder_input_size=2 * hidden_size if direction_option == 'bi_sep' else hidden_size,
                                         decoder_hidden_size=hidden_size,
                                         word_emb=self.word_emb, vocab=self.vocab.in_word_vocab,
                                         attention_type="sep_diff_encoder_type", fuse_strategy="concatenate",
                                         rnn_emb_input_size=hidden_size,
                                         tgt_emb_as_output_layer=True, device=self.graph_topology.device).to(device)
        self.loss_calc = Graph2seqLoss(self.vocab.in_word_vocab)

    def forward(self, graph_list, tgt=None, require_loss=True):
        batch_graph = self.graph_topology(graph_list)
        batch_graph = self.gnn_encoder(batch_graph)
        batch_graph.node_features['rnn_emb'] = batch_graph.node_features['node_feat']

        # down-task
        prob, _, _ = self.seq_decoder(from_batch(batch_graph), tgt_seq=tgt)
        if require_loss:
            loss = self.loss_calc(prob, tgt)
            return prob, loss
        else:
            return prob


class Jobs:
    def __init__(self, args):
        super(Jobs, self).__init__()
        self.device = args.device
        self._build_dataloader()
        self._build_model()
        self._build_optimizer()
        self._build_evaluation()

    def _build_dataloader(self):
        dataset = JobsDataset(root_dir="graph4nlp/pytorch/test/dataset/jobs",
                              topology_builder=DependencyBasedGraphConstruction,
                              topology_subdir='DependencyGraph', share_vocab=True, val_split_ratio=0.2)

        self.train_dataloader = DataLoader(dataset.train, batch_size=24, shuffle=True, num_workers=1,
                                           collate_fn=dataset.collate_fn)
        self.val_dataloader = DataLoader(dataset.val, batch_size=24, shuffle=True, num_workers=1,
                                         collate_fn=dataset.collate_fn)
        self.test_dataloader = DataLoader(dataset.test, batch_size=24, shuffle=True, num_workers=1,
                                          collate_fn=dataset.collate_fn)
        self.vocab = dataset.vocab_model

    def _build_model(self):
        self.model = Graph2seq(self.vocab, device=self.device)

    def _build_optimizer(self):
        parameters = [p for p in self.model.parameters() if p.requires_grad]
        self.optimizer = optim.Adam(parameters, lr=1e-3)

    def _build_evaluation(self):
        self.metrics = [ExpressionAccuracy()]

    def train(self):
        max_score = -1
        for epoch in range(200):
            self.model.train()
            print("Epoch: {}".format(epoch))
            for data in self.train_dataloader:
                graph_list, tgt = data
                tgt = tgt.to(self.device)
                _, loss = self.model(graph_list, tgt, require_loss=True)
                print(loss)
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()
            if epoch > 4:
                score = self.evaluate(self.val_dataloader)
                if score > max_score:
                    torch.save(self.model.state_dict(), 'best_model.pt')
                max_score = max(max_score, score)
        return max_score

    def evaluate(self, dataloader):
        self.model.eval()
        pred_collect = []
        gt_collect = []
        for data in dataloader:
            graph_list, tgt = data
            prob = self.model(graph_list, require_loss=False)
            pred = logits2seq(prob)

            pred_str = wordid2str(pred.detach().cpu(), self.vocab.in_word_vocab)
            tgt_str = wordid2str(tgt, self.vocab.in_word_vocab)
            pred_collect.extend(pred_str)
            gt_collect.extend(tgt_str)

        score = self.metrics[0].calculate_scores(ground_truth=gt_collect, predict=pred_collect)
        print("accuracy: {:.3f}".format(score))
        return score


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--device', type=str, default='cpu')
    args = parser.parse_args()
    runner = Jobs(args)
    max_score = runner.train()
    print("Train finish, best score: {:.3f}".format(max_score))
    runner.model.load_state_dict(torch.load('best_model.pt'))
    test_score = runner.evaluate(runner.test_dataloader)
    print("Test score: {:.3f}".format(test_score))



import torch
from torch import nn
from torch.nn.utils.rnn import pad_packed_sequence, pack_padded_sequence
import dgl

from ..utils.generic_utils import to_cuda, dropout_fn, create_mask
from ..utils.bert_utils import *


class EmbeddingConstructionBase(nn.Module):
    """
    Base class for (initial) graph embedding construction.

    ...

    Attributes
    ----------
    feat : dict
        Raw features of graph nodes and/or edges.

    Methods
    -------
    forward(raw_text_data)
        Generate dynamic graph topology and embeddings.
    """

    def __init__(self):
        super(EmbeddingConstructionBase, self).__init__()

    def forward(self):
        raise NotImplementedError()

class EmbeddingConstructionBase(nn.Module):
    """Basic class for embedding construction.
    """
    def __init__(self):
        super(EmbeddingConstructionBase, self).__init__()

    def forward(self):
        """Compute initial node/edge embeddings.

        Raises
        ------
        NotImplementedError
            NotImplementedError.
        """
        raise NotImplementedError()

class EmbeddingConstruction(EmbeddingConstructionBase):
    """Initial graph embedding construction class.

    Parameters
    ----------
    word_vocab : Vocab
        The word vocabulary.
    single_token_item : bool
        Specify whether the item (i.e., node or edge) contains single token or multiple tokens.
    emb_strategy : str
        Specify the embedding construction strategy including the following options:
            - 'w2v': use word2vec embeddings.
            - 'w2v_bilstm': use word2vec embeddings, and apply BiLSTM encoders.
            - 'w2v_bigru': use word2vec embeddings, and apply BiGRU encoders.
            - 'bert': use BERT embeddings.
            - 'bert_bilstm': use BERT embeddings, and apply BiLSTM encoders.
            - 'bert_bigru': use BERT embeddings, and apply BiGRU encoders.
            - 'w2v_bert': use word2vec and BERT embeddings.
            - 'w2v_bert_bilstm': use word2vec and BERT embeddings, and apply BiLSTM encoders.
            - 'w2v_bert_bigru': use word2vec and BERT embeddings, and apply BiGRU encoders.
    hidden_size : int, optional
        The hidden size of RNN layer, default: ``None``.
    num_rnn_layers : int, optional
        The number of RNN layers, default: ``1``.
    fix_word_emb : boolean, optional
        Specify whether to fix pretrained word embeddings, default: ``True``.
    fix_bert_emb : boolean, optional
        Specify whether to fix pretrained BERT embeddings, default: ``True``.
    bert_model_name : str, optional
        Specify the BERT model name, default: ``'bert-base-uncased'``.
    bert_lower_case : bool, optional
        Specify whether to lower case the input text for BERT embeddings, default: ``True``.
    word_dropout : float, optional
        Dropout ratio for word embedding, default: ``None``.
    rnn_dropout : float, optional
        Dropout ratio for RNN embedding, default: ``None``.
    device : torch.device, optional
        Specify computation device (e.g., CPU), default: ``None`` for using CPU.

    Note
    ----------
        word_emb_type : str or list of str
            Specify pretrained word embedding types including "w2v", "node_edge_bert", or "seq_bert".
        node_edge_emb_strategy : str
            Specify node/edge embedding strategies including "mean", "bilstm" and "bigru".
        seq_info_encode_strategy : str
            Specify strategies of encoding sequential information in raw text
            data including "none", "bilstm" and "bigru". You might
            want to do this in some situations, e.g., when all the nodes are single
            tokens extracted from the raw text.

        1) single-token node (i.e., single_token_item=`True`):
            a) 'w2v', 'bert', 'w2v_bert'
            b) node_edge_emb_strategy: 'mean'
            c) seq_info_encode_strategy: 'none', 'bilstm', 'bigru'
            emb_strategy: 'w2v', 'w2v_bilstm', 'w2v_bigru',
            'bert', 'bert_bilstm', 'bert_bigru',
            'w2v_bert', 'w2v_bert_bilstm', 'w2v_bert_bigru'

        2) multi-token node (i.e., single_token_item=`False`):
            a) 'w2v', 'bert', 'w2v_bert'
            b) node_edge_emb_strategy: 'mean', 'bilstm', 'bigru'
            c) seq_info_encode_strategy: 'none'
            emb_strategy: ('w2v', 'w2v_bilstm', 'w2v_bigru',
            'bert', 'bert_bilstm', 'bert_bigru',
            'w2v_bert', 'w2v_bert_bilstm', 'w2v_bert_bigru')
    """
    def __init__(self,
                    word_vocab,
                    single_token_item,
                    emb_strategy='w2v_bilstm',
                    hidden_size=None,
                    num_rnn_layers=1,
                    fix_word_emb=True,
                    fix_bert_emb=True,
                    bert_model_name='bert-base-uncased',
                    bert_lower_case=True,
                    word_dropout=None,
                    rnn_dropout=None,
                    device=None):
        super(EmbeddingConstruction, self).__init__()
        self.device = device
        self.word_dropout = word_dropout
        self.rnn_dropout = rnn_dropout

        assert emb_strategy in ('w2v', 'w2v_bilstm', 'w2v_bigru',
                        'bert', 'bert_bilstm', 'bert_bigru',
                        'w2v_bert', 'w2v_bert_bilstm', 'w2v_bert_bigru'),\
            "emb_strategy must be one of ('w2v', 'w2v_bilstm', 'w2v_bigru', 'bert', 'bert_bilstm', 'bert_bigru', 'w2v_bert', 'w2v_bert_bilstm', 'w2v_bert_bigru')"

        word_emb_type = set()
        if single_token_item:
            node_edge_emb_strategy = 'mean'
            if 'w2v' in emb_strategy:
                word_emb_type.add('w2v')

            if 'bert' in emb_strategy:
                word_emb_type.add('seq_bert')

            if 'bilstm' in emb_strategy:
                seq_info_encode_strategy = 'bilstm'
            elif 'bigru' in emb_strategy:
                seq_info_encode_strategy = 'bigru'
            else:
                seq_info_encode_strategy = 'none'
        else:
            seq_info_encode_strategy = 'none'
            if 'w2v' in emb_strategy:
                word_emb_type.add('w2v')

            if 'bert' in emb_strategy:
                word_emb_type.add('node_edge_bert')

            if 'bilstm' in emb_strategy:
                node_edge_emb_strategy = 'bilstm'
            elif 'bigru' in emb_strategy:
                node_edge_emb_strategy = 'bigru'
            else:
                node_edge_emb_strategy = 'mean'


        word_emb_size = 0
        self.word_emb_layers = nn.ModuleDict()
        if 'w2v' in word_emb_type:
            self.word_emb_layers['w2v'] = WordEmbedding(
                            word_vocab.embeddings.shape[0],
                            word_vocab.embeddings.shape[1],
                            pretrained_word_emb=word_vocab.embeddings,
                            fix_emb=fix_word_emb,
                            device=self.device)
            word_emb_size += word_vocab.embeddings.shape[1]

        if 'node_edge_bert' in word_emb_type:
            self.word_emb_layers['node_edge_bert'] = BertEmbedding(name=bert_model_name,
                                                            fix_emb=fix_bert_emb,
                                                            lower_case=bert_lower_case)
            word_emb_size += self.word_emb_layers['node_edge_bert'].bert_model.config.hidden_size

        if 'seq_bert' in word_emb_type:
            self.word_emb_layers['seq_bert'] = BertEmbedding(name=bert_model_name,
                                                            fix_emb=fix_bert_emb,
                                                            lower_case=bert_lower_case)

        if node_edge_emb_strategy in ('bilstm', 'bigru'):
            self.node_edge_emb_layer = RNNEmbedding(
                                    word_emb_size,
                                    hidden_size,
                                    bidirectional=True,
                                    num_layers=num_rnn_layers,
                                    rnn_type='lstm' if node_edge_emb_strategy == 'bilstm' else 'gru',
                                    dropout=rnn_dropout,
                                    device=device)
            rnn_input_size = hidden_size
        else:
            self.node_edge_emb_layer = MeanEmbedding()
            rnn_input_size = word_emb_size

        if 'seq_bert' in word_emb_type:
            rnn_input_size += self.word_emb_layers['seq_bert'].bert_model.config.hidden_size

        if seq_info_encode_strategy in ('bilstm', 'bigru'):
            self.output_size = hidden_size
            self.seq_info_encode_layer = RNNEmbedding(
                                    rnn_input_size,
                                    hidden_size,
                                    bidirectional=True,
                                    num_layers=num_rnn_layers,
                                    rnn_type='lstm' if seq_info_encode_strategy == 'bilstm' else 'gru',
                                    dropout=rnn_dropout,
                                    device=device)

            # apply a linear projection to make rnn_input_size equal hidden_size
            if rnn_input_size != hidden_size:
                self.linear_transform = nn.Linear(rnn_input_size, hidden_size, bias=False)

        else:
            self.output_size = rnn_input_size
            self.seq_info_encode_layer = None


    def forward(self, batch_gd, item_size, num_items, num_word_items=None):
        """Compute initial node/edge embeddings.

        Parameters
        ----------
        batch_gd : GraphData
            The batched graph data.
        item_size : torch.LongTensor
            The length of word sequence per item with shape :math:`(N)`.
        num_items : torch.LongTensor
            The number of items per graph with shape :math:`(B,)`
            where :math:`B` is the number of graphs in the batched graph.
        num_word_items : torch.LongTensor, optional
            The number of word items (that are extracted from the raw text)
            per graph with shape :math:`(B,)` where :math:`B` is the number
            of graphs in the batched graph. We assume that the word items are
            not reordered and interpolated, and always appear before the non-word
            items in the graph. Default: ``None``.

        Returns
        -------
        torch.Tensor
            The output item embeddings.
        """

        feat = []
        if 'w2v' in self.word_emb_layers:
            input_data = batch_gd.node_features['token_id'].long()
            feat.append(self.word_emb_layers['w2v'](input_data))

        if 'node_edge_bert' in self.word_emb_layers:
            input_data = [[batch_gd.node_attributes[i]['token']] for i in range(batch_gd.get_node_num())]
            feat.append(self.word_emb_layers['node_edge_bert'](input_data))

        if len(feat) > 0:
            feat = torch.cat(feat, dim=-1)
            feat = dropout_fn(feat, self.word_dropout, shared_axes=[-2], training=self.training)
            feat = self.node_edge_emb_layer(feat, item_size)
            if isinstance(feat, (tuple, list)):
                feat = feat[-1]


        if self.seq_info_encode_layer is None and 'seq_bert' not in self.word_emb_layers:
            return feat
        else:
            # unbatching
            new_feat = []
            raw_text_data = []
            start_idx = 0
            max_num_items = torch.max(num_items).item()
            for i in range(num_items.shape[0]):
                if len(feat) > 0:
                    tmp_feat = feat[start_idx: start_idx + num_items[i].item()]
                    if num_items[i].item() < max_num_items:
                        tmp_feat = torch.cat([tmp_feat, to_cuda(torch.zeros(
                            max_num_items - num_items[i], tmp_feat.shape[1]), self.device)], 0)
                    new_feat.append(tmp_feat)

                if 'seq_bert' in self.word_emb_layers:
                    raw_text_data.append([batch_gd.node_attributes[j]['token'] for j in range(start_idx, start_idx + num_items[i].item())])

                start_idx += num_items[i].item()

            # computation
            if len(new_feat) > 0:
                new_feat = torch.stack(new_feat, 0)

            if 'seq_bert' in self.word_emb_layers:
                bert_feat = self.word_emb_layers['seq_bert'](raw_text_data)
                if len(new_feat) > 0:
                    new_feat = torch.cat([new_feat, bert_feat], -1)
                else:
                    new_feat = bert_feat


            if self.seq_info_encode_layer is None:
                return new_feat

            len_ = num_word_items if num_word_items is not None else num_items
            rnn_state = self.seq_info_encode_layer(new_feat, len_)
            if isinstance(rnn_state, (tuple, list)):
                rnn_state = rnn_state[0]

            # batching
            ret_feat = []
            for i in range(len_.shape[0]):
                tmp_feat = rnn_state[i][:len_[i]]
                if len(tmp_feat) < num_items[i].item():
                    prev_feat = new_feat[i, len_[i]: num_items[i]]
                    if prev_feat.shape[-1] != tmp_feat.shape[-1]:
                        prev_feat = self.linear_transform(prev_feat)

                    tmp_feat = torch.cat([tmp_feat, prev_feat], 0)
                ret_feat.append(tmp_feat)

            ret_feat = torch.cat(ret_feat, 0)

            return ret_feat


class WordEmbedding(nn.Module):
    """Word embedding class.

    Parameters
    ----------
    vocab_size : int
        The word vocabulary size.
    emb_size : int
        The word embedding size.
    padding_idx : int, optional
        The padding index, default: ``0``.
    pretrained_word_emb : numpy.ndarray, optional
        The pretrained word embeddings, default: ``None``.
    fix_emb : boolean, optional
        Specify whether to fix pretrained word embeddings, default: ``True``.

    Examples
    ----------
    >>> word_emb_layer = WordEmbedding(1000, 300, padding_idx=0, pretrained_word_emb=None, fix_emb=True)
    """
    def __init__(self, vocab_size, emb_size, padding_idx=0,
                    pretrained_word_emb=None, fix_emb=True, device=None):
        super(WordEmbedding, self).__init__()
        self.word_emb_layer = nn.Embedding(vocab_size, emb_size, padding_idx=padding_idx,
                            _weight=torch.from_numpy(pretrained_word_emb).float()
                            if pretrained_word_emb is not None else None)
        self.device = device

        if fix_emb:
            print('[ Fix word embeddings ]')
            for param in self.word_emb_layer.parameters():
                param.requires_grad = False

    def forward(self, input_tensor):
        """Compute word embeddings.

        Parameters
        ----------
        input_tensor : torch.LongTensor
            The input word index sequence, shape: [num_items, max_size].

        Returns
        -------
        torch.Tensor
            Word embedding matrix.
        """
        return self.word_emb_layer(input_tensor)

class BertEmbedding(nn.Module):
    """Bert embedding class.

    Parameters
    ----------
    name : str, optional
        BERT model name, default: ``'bert-base-uncased'``.
    max_seq_len : int, optional
        Maximal sequence length, default: ``500``.
    doc_stride : int, optional
        Chunking stride, default: ``250``.
    fix_emb : boolean, optional
        Specify whether to fix pretrained BERT embeddings, default: ``True``.
    lower_case : boolean, optional
        Specify whether to use lower case, default: ``True``.

    """
    def __init__(self,
                name='bert-base-uncased',
                max_seq_len=500,
                doc_stride=250,
                fix_emb=True,
                lower_case=True):
        super(BertEmbedding, self).__init__()
        self.bert_max_seq_len = max_seq_len
        self.bert_doc_stride = doc_stride
        self.fix_emb = fix_emb

        from transformers import BertModel
        from transformers import BertTokenizer
        print('[ Using pretrained BERT embeddings ]')
        self.bert_tokenizer = BertTokenizer.from_pretrained(name, do_lower_case=lower_case)
        self.bert_model = BertModel.from_pretrained(name)
        if fix_emb:
            print('[ Fix BERT layers ]')
            self.bert_model.eval()
            for param in self.bert_model.parameters():
                param.requires_grad = False
        else:
            print('[ Finetune BERT layers ]')
            self.bert_model.train()

        # compute weighted average over BERT layers
        self.logits_bert_layers = nn.Parameter(nn.init.xavier_uniform_(torch.Tensor(1, self.bert_model.config.num_hidden_layers)))


    def forward(self, raw_text_data):
        """Compute BERT embeddings for each word in text.

        Parameters
        ----------
        raw_text_data : list
            The raw text input data. Example: [['what', 'is', 'bert'], ['how', 'to', 'use', 'bert']].

        Returns
        -------
        torch.Tensor
            BERT embedding matrix.
        """
        bert_features = []
        max_d_len = 0
        for text in raw_text_data:
            bert_features.append(convert_text_to_bert_features(text, self.bert_tokenizer, self.bert_max_seq_len, self.bert_doc_stride))
            max_d_len = max(max_d_len, len(text))

        max_bert_d_num_chunks = max([len(ex_bert_d) for ex_bert_d in bert_features])
        max_bert_d_len = max([len(bert_d.input_ids) for ex_bert_d in bert_features for bert_d in ex_bert_d])
        bert_xd = torch.LongTensor(len(raw_text_data), max_bert_d_num_chunks, max_bert_d_len).fill_(0)
        bert_xd_mask = torch.LongTensor(len(raw_text_data), max_bert_d_num_chunks, max_bert_d_len).fill_(0)
        for i, ex_bert_d in enumerate(bert_features): # Example level
            for j, bert_d in enumerate(ex_bert_d): # Chunk level
                bert_xd[i, j, :len(bert_d.input_ids)].copy_(torch.LongTensor(bert_d.input_ids))
                bert_xd_mask[i, j, :len(bert_d.input_mask)].copy_(torch.LongTensor(bert_d.input_mask))

        bert_xd = bert_xd.to(self.bert_model.device)
        bert_xd_mask = bert_xd_mask.to(self.bert_model.device)

        encoder_outputs = self.bert_model(bert_xd.view(-1, bert_xd.size(-1)),
                                        token_type_ids=None,
                                        attention_mask=bert_xd_mask.view(-1, bert_xd_mask.size(-1)),
                                        output_hidden_states=True,
                                        return_dict=True)
        all_encoder_layers = encoder_outputs['hidden_states'][1:] # The first one is the input embedding
        all_encoder_layers = torch.stack([x.view(bert_xd.shape + (-1,)) for x in all_encoder_layers], 0)
        bert_xd_f = extract_bert_hidden_states(all_encoder_layers, max_d_len, bert_features, weighted_avg=True)

        weights_bert_layers = torch.softmax(self.logits_bert_layers, dim=-1)
        bert_xd_f = torch.mm(weights_bert_layers, bert_xd_f.view(bert_xd_f.size(0), -1)).view(bert_xd_f.shape[1:])

        return bert_xd_f



class MeanEmbedding(nn.Module):
    """Mean embedding class.
    """
    def __init__(self):
        super(MeanEmbedding, self).__init__()

    def forward(self, emb, len_):
        """Compute average embeddings.

        Parameters
        ----------
        emb : torch.Tensor
            The input embedding tensor.
        len_ : torch.Tensor
            The sequence length tensor.

        Returns
        -------
        torch.Tensor
            The average embedding tensor.
        """
        sumed_emb = torch.sum(emb, dim=1)
        len_ = len_.unsqueeze(1).expand_as(sumed_emb).float()
        return sumed_emb / len_


class RNNEmbedding(nn.Module):
    """RNN embedding class: apply the RNN network to a sequence of word embeddings.

    Parameters
    ----------
    input_size : int
        The input feature size.
    hidden_size : int
        The hidden layer size.
    dropout : float, optional
        Dropout ratio, default: ``None``.
    bidirectional : boolean, optional
        Whether to use bidirectional RNN, default: ``False``.
    rnn_type : str
        The RNN cell type, default: ``lstm``.
    device : torch.device, optional
        Specify computation device (e.g., CPU), default: ``None`` for using CPU.
    """
    def __init__(self,
                input_size,
                hidden_size,
                bidirectional=False,
                num_layers=1,
                rnn_type='lstm',
                dropout=None,
                device=None):
        super(RNNEmbedding, self).__init__()
        if not rnn_type in ('lstm', 'gru'):
            raise RuntimeError('rnn_type is expected to be lstm or gru, got {}'.format(rnn_type))

        # if bidirectional:
        #     print('[ Using bidirectional {} encoder ]'.format(rnn_type))
        # else:
        #     print('[ Using {} encoder ]'.format(rnn_type))

        if bidirectional and hidden_size % 2 != 0:
            raise RuntimeError('hidden_size is expected to be even in the bidirectional mode!')

        self.device = device
        self.dropout = dropout
        self.rnn_type = rnn_type
        self.num_layers = num_layers
        self.num_directions = 2 if bidirectional else 1
        self.hidden_size = hidden_size // 2 if bidirectional else hidden_size
        model = nn.LSTM if rnn_type == 'lstm' else nn.GRU
        self.model = model(input_size, self.hidden_size, num_layers, batch_first=True, bidirectional=bidirectional)

    def forward(self, x, x_len):
        """Apply the RNN network to a sequence of word embeddings.

        Parameters
        ----------
        x : torch.Tensor
            The word embedding sequence.
        x_len : torch.LongTensor
            The input sequence length.

        Returns
        -------
        torch.Tensor
            The hidden states at every time step.
        torch.Tensor
            The hidden state at the last time step.
        """
        sorted_x_len, indx = torch.sort(x_len, 0, descending=True)
        x = pack_padded_sequence(x[indx], sorted_x_len.data.tolist(), batch_first=True)

        h0 = to_cuda(torch.zeros(self.num_directions * self.num_layers, x_len.size(0), self.hidden_size), self.device)
        if self.rnn_type == 'lstm':
            c0 = to_cuda(torch.zeros(self.num_directions * self.num_layers, x_len.size(0), self.hidden_size), self.device)
            packed_h, (packed_h_t, _) = self.model(x, (h0, c0))
        else:
            packed_h, packed_h_t = self.model(x, h0)

        if self.num_layers > 1:
            # use the last RNN layer hidden state
            packed_h_t = packed_h_t.view(self.num_layers, self.num_directions, -1, self.hidden_size)[-1]

        if self.num_directions == 2:
            packed_h_t = torch.cat((packed_h_t[-1], packed_h_t[-2]), 1)
        else:
            packed_h_t = packed_h_t[-1]

        hh, _ = pad_packed_sequence(packed_h, batch_first=True)

        # restore the sorting
        _, inverse_indx = torch.sort(indx, 0)
        restore_hh = hh[inverse_indx]
        restore_packed_h_t = packed_h_t[inverse_indx]

        # add dropout
        restore_hh = dropout_fn(restore_hh, self.dropout, shared_axes=[-2], training=self.training)
        restore_packed_h_t = dropout_fn(restore_packed_h_t, self.dropout, training=self.training)

        return restore_hh, restore_packed_h_t



from nltk.tokenize import word_tokenize
import torch
from torch import nn

from ...data.data import GraphData
from .base import DynamicGraphConstructionBase
from ..utils.generic_utils import normalize_adj, to_cuda
from ..utils.constants import VERY_SMALL_NUMBER
from .utils import convert_adj_to_graph
from ...data.data import to_batch


class NodeEmbeddingBasedGraphConstruction(DynamicGraphConstructionBase):
    """Class for node embedding based dynamic graph construction.

    Parameters
    ----------
    word_vocab : Vocab
        The word vocabulary.
    embedding_styles : dict
        - ``word_emb_type`` : Specify pretrained word embedding types
            including "w2v" and/or "bert".
        - ``node_edge_emb_strategy`` : Specify node/edge embedding
            strategies including "mean", "lstm", "gru", "bilstm" and "bigru".
        - ``seq_info_encode_strategy`` : Specify strategies of encoding
            sequential information in raw text data including "none",
            "lstm", "gru", "bilstm" and "bigru".
    """
    def __init__(self, word_vocab, embedding_styles, **kwargs):
        super(NodeEmbeddingBasedGraphConstruction, self).__init__(
                                                            word_vocab,
                                                            embedding_styles,
                                                            **kwargs)

    def forward(self, batch_graphdata: list):
        """Compute graph topology and initial node embeddings.

        Parameters
        ----------
        batch_graphdata : list of GraphData
            The input graph data list.
        Returns
        -------
        GraphData
            The constructed batched graph.
        """
        node_size = []
        num_nodes = []

        for g in batch_graphdata:
            g.node_features['token_id'] = to_cuda(g.node_features['token_id'], self.device)
            num_nodes.append(g.get_node_num())
            node_size.extend([1 for i in range(num_nodes[-1])])

        node_size = to_cuda(torch.Tensor(node_size), self.device).int()
        num_nodes = to_cuda(torch.Tensor(num_nodes), self.device).int()
        batch_gd = to_batch(batch_graphdata)

        node_emb = self.embedding(batch_gd, node_size, num_nodes)

        node_mask = self._get_node_mask_for_batch_graph(num_nodes)
        new_batch_gd = self.topology(node_emb, node_mask=node_mask)
        new_batch_gd.node_features['node_feat'] = node_emb
        # new_batch_gd.batch = batch_gd.batch
        new_batch_gd.copy_batch_info(batch_gd)

        return new_batch_gd


    def topology(self, node_emb, node_mask=None):
        """Compute graph topology.

        Parameters
        ----------
        node_emb : torch.Tensor
            The node embeddings.
        node_mask : torch.Tensor, optional
            The node mask matrix, default: ``None``.

        Returns
        -------
        GraphData
            The constructed graph.
        """
        raw_adj = self.compute_similarity_metric(node_emb, node_mask)
        raw_adj = self.sparsify_graph(raw_adj)
        graph_reg = self.compute_graph_regularization(raw_adj, node_emb)

        if self.sim_metric_type in ('rbf_kernel', 'weighted_cosine'):
            assert raw_adj.min().item() >= 0, 'adjacency matrix must be non-negative!'
            adj = raw_adj / torch.clamp(torch.sum(raw_adj, dim=-1, keepdim=True), min=VERY_SMALL_NUMBER)
            reverse_adj = raw_adj / torch.clamp(torch.sum(raw_adj, dim=0, keepdim=True), min=VERY_SMALL_NUMBER)
        elif self.sim_metric_type == 'cosine':
            raw_adj = (raw_adj > 0).float()
            adj = normalize_adj(raw_adj)
            reverse_adj = adj
        else:
            adj = torch.softmax(raw_adj, dim=-1)
            reverse_adj = torch.softmax(raw_adj, dim=0)

        graph_data = convert_adj_to_graph(adj, reverse_adj, 0)
        graph_data.graph_attributes['graph_reg'] = graph_reg

        return graph_data


    def embedding(self, node_word_idx, node_size, num_nodes):
        """Compute initial node embeddings.

        Parameters
        ----------
        node_word_idx : torch.LongTensor
            The input word index node features.
        node_size : torch.LongTensor
            Indicate the length of word sequences for nodes.
        num_nodes : torch.LongTensor
            Indicate the number of nodes.

        Returns
        -------
        torch.Tensor
            The initial node embeddings.
        """
        return self.embedding_layer(node_word_idx, node_size, num_nodes)


    @classmethod
    def init_topology(cls, raw_text_data, lower_case=True, tokenizer=word_tokenize):
        """Convert raw text data to the initial node set graph (i.e., no edge information).

        Parameters
        ----------
        raw_text_data : str or list/tuple of str
            The raw text data. When a list/tuple of tokens is provided, no
            tokenization will be conducted and each token is a node;
            otherwise, tokenization will be conducted on the input string
            to get a list of tokens.
        lower_case : boolean
            Specify whether to lower case the input text, default: ``True``.
        tokenizer : callable, optional
            The tokenization function.

        Returns
        -------
        GraphData
            The constructed graph.
        """
        if isinstance(raw_text_data, str):
            token_list = tokenizer(raw_text_data.strip())
        elif isinstance(raw_text_data, (list, tuple)):
            token_list = raw_text_data
        else:
            raise RuntimeError('raw_text_data must be str or list/tuple of str')

        graph = GraphData()
        graph.add_nodes(len(token_list))

        for idx in range(len(token_list)):
            graph.node_attributes[idx]['token'] = token_list[idx].lower() if lower_case else token_list[idx]

        return graph


import os
import argparse
import numpy as np
from scipy import sparse
import networkx as nx
import time
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.backends.cudnn as cudnn
import dgl
from dgl import DGLGraph
import dgl.function as fn
from dgl.data import register_data_args, load_data

from ..utils import EarlyStopping
from ...modules.graph_construction import NodeEmbeddingBasedGraphConstruction, NodeEmbeddingBasedRefinedGraphConstruction
from ...modules.utils.vocab_utils import VocabModel
from ...modules.utils.padding_utils import pad_2d_vals_no_size
from ...modules.utils.generic_utils import to_cuda


def main(args, seed):
    # Configure
    np.random.seed(seed)
    torch.manual_seed(seed)

    if not args.no_cuda and torch.cuda.is_available():
        print('[ Using CUDA ]')
        device = torch.device('cuda' if args.gpu < 0 else 'cuda:%d' % args.gpu)
        cudnn.benchmark = True
        torch.cuda.manual_seed(seed)
    else:
        device = torch.device('cpu')


    raw_text_data = [['I like nlp.', 'Same here!'], ['I like graph.', 'Same here!']]

    vocab_model = VocabModel(raw_text_data, max_word_vocab_size=None,
                                min_word_vocab_freq=1,
                                word_emb_size=300)

    src_text_seq = list(zip(*raw_text_data))[0]

    # Test raw_text_to_init_graph method of dynamic graph construction class
    text_graphs = []
    for each in src_text_seq:
        tmp_graph = NodeEmbeddingBasedGraphConstruction.raw_text_to_init_graph(each)
        text_graphs.append(tmp_graph)

    src_idx_seq = [vocab_model.word_vocab.to_index_sequence(each) for each in src_text_seq]
    src_len = torch.LongTensor([len(each) for each in src_idx_seq]).to(device)
    num_seq = torch.LongTensor([len(src_len)]).to(device)
    input_tensor = torch.LongTensor(pad_2d_vals_no_size(src_idx_seq)).to(device)
    print('input_tensor: {}'.format(input_tensor.shape))


    embedding_styles = {'word_emb_type': 'w2v',
                        'node_edge_emb_strategy': 'bilstm',
                        'seq_info_encode_strategy': 'none'}

    gl = NodeEmbeddingBasedGraphConstruction(
                                            vocab_model.word_vocab,
                                            embedding_styles,
                                            input_size=args.gl_num_hidden,
                                            hidden_size=args.gl_num_hidden,
                                            top_k_neigh=args.gl_topk,
                                            device=device)

    gl.to(device)
    print(gl)

    graph = gl(input_tensor, src_len, num_seq)
    print('\nadj:', graph.adjacency_matrix().to_dense())
    print('node feat: ', graph.ndata['node_feat'])


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='DynamicGraphConstruction')
    register_data_args(parser)
    parser.add_argument("--no-cuda", action="store_true", default=False,
                        help="use CPU")
    parser.add_argument("--gpu", type=int, default=-1,
                        help="which GPU to use.")
    parser.add_argument("--epochs", type=int, default=1000,
                        help="number of training epochs")
    parser.add_argument("--gl-num-hidden", type=int, default=16,
                        help="number of hidden units for dynamic graph construction")
    parser.add_argument("--gl-topk", type=int, default=200,
                        help="top k for dynamic graph construction")
    parser.add_argument("--gl-type", type=str, default='node_emb',
                        help=r"dynamic graph construction algorithm type, \
                        {'node_emb', 'node_edge_emb' and 'node_emb_refined'},\
                        default: 'node_emb'")
    parser.add_argument("--init-adj-alpha", type=float, default=0.8,
                        help="alpha ratio for combining initial graph adjacency matrix")
    parser.add_argument("--num-hidden", type=int, default=16,
                        help="number of hidden units")
    parser.add_argument("--lr", type=float, default=0.01,
                        help="learning rate")
    parser.add_argument('--weight-decay', type=float, default=5e-4,
                        help="weight decay")
    parser.add_argument('--early-stop', action='store_true', default=False,
                        help="indicates whether to use early stop or not")
    parser.add_argument("--patience", type=int, default=100,
                        help="early stopping patience")
    parser.add_argument('--fastmode', action="store_true", default=False,
                        help="skip re-evaluate the validation set")
    parser.add_argument('--save-model-path', type=str, default="checkpoint",
                        help="path to the best saved model")
    args = parser.parse_args()
    args.save_model_path = '{}_{}_gl_type_{}_gl_topk_{}_init_adj_alpha_{}'.format(
                                                    args.save_model_path,
                                                    args.dataset,
                                                    args.gl_type,
                                                    args.gl_topk,
                                                    args.init_adj_alpha)
    print(args)
    main(args, 1234)



import torchtext.vocab as vocab
from graph4nlp.pytorch.modules.utils.vocab_utils import Vocab
import torch, logging


def get_log(log_file):
    logger = logging.getLogger(log_file)
    logger.setLevel(logging.DEBUG)
    fh = logging.FileHandler(log_file)
    fh.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    ch.setFormatter(formatter)
    fh.setFormatter(formatter)
    logger.addHandler(ch)
    logger.addHandler(fh)
    return logger

def get_glove_weights(vocabulary: Vocab, dim=300):
    print(vocabulary.index2word)

    glove = vocab.GloVe(name='6B', dim=dim)
    vocab_size = vocabulary.get_vocab_size()
    weight = torch.randn(vocab_size, dim)
    for i, word in enumerate(vocabulary.index2word):
        glove_index = glove.stoi.get(word)
        glove_rep = glove.get_vecs_by_tokens(word)
        weight[i, :] = glove_rep
    return weight


def wordid2str(word_ids, vocab: Vocab):
    ret = []
    assert len(word_ids.shape) == 2, print(word_ids.shape)
    for i in range(word_ids.shape[0]):
        id_list = word_ids[i, :]
        ret_inst = []
        for j in range(id_list.shape[0]):
            if id_list[j] == vocab.EOS:
                break
            token = vocab.getWord(id_list[j])
            ret_inst.append(token)
        ret.append(" ".join(ret_inst))
    return ret



"""
The Graph4NLP library uses the class :py:class:`GraphData` as the representation for structured data (graphs).
:py:class:`GraphData` supports basic operations to the graph, including adding nodes and edges. :py:class:`GraphData` also
supports adding features which are in tensor form, and attributes which are of arbitrary form to the corresponding
nodes or edges. Batching operations is also supported by :py:class:`GraphData`.

"""
import warnings
from collections import namedtuple

import dgl
import numpy as np
import scipy.sparse
import torch

from .utils import SizeMismatchException, EdgeNotFoundException
from .utils import check_and_expand, int_to_list, entail_zero_padding, slice_to_list, reverse_index
from .views import NodeView, NodeFeatView, EdgeView

EdgeIndex = namedtuple('EdgeIndex', ['src', 'tgt'])

node_feat_factory = dict
node_attr_factory = dict
single_node_attr_factory = dict
res_init_node_attr = {'node_attr': None}
res_init_node_features = {'node_feat': None, 'node_emb': None}

eid_nids_mapping_factory = dict
nids_eid_mapping_factory = dict
edge_feature_factory = dict
edge_attribute_factory = dict
single_edge_attr_factory = dict
res_init_edge_features = {'edge_feat': None, 'edge_emb': None, 'edge_weight': None}
res_init_edge_attributes = {'edge_attr': None}

graph_data_factory = dict


class GraphData(object):
    """
    Represent a single graph with additional attributes.
    """

    def __init__(self, src=None, device=None):
        """
        Parameters
        ----------
        src: GraphData, default=None
            The source graph. If not None, then the newly generated graph is a copy of :py:class:`src`.
        device: str, default=None
            The device descriptor for graph. By default it is None.
        """

        # Initialize internal data storages.
        self._node_attributes = node_attr_factory()
        self._node_features = node_feat_factory(res_init_node_features)
        self._edge_indices = EdgeIndex(src=[], tgt=[])
        self._eid_nids_mapping = eid_nids_mapping_factory()
        self._nids_eid_mapping = nids_eid_mapping_factory()
        self._edge_features = edge_feature_factory(res_init_edge_features)
        self._edge_attributes = edge_attribute_factory()
        self.graph_attributes = graph_data_factory()
        self.device = device

        # Batch information. If this instance is not a batch, then the following attributes are all `None`.
        self.batch = None
        self.batch_size = None
        self._batch_num_nodes = None
        self._batch_num_edges = None

        if src is not None:
            if isinstance(src, GraphData):
                self.from_graphdata(src)
            else:
                raise NotImplementedError

    def to(self, device):
        self.device = device
        return self

    # Node operations
    @property
    def nodes(self) -> NodeView:
        """
        Return a node view through which the user can access the features and attributes

        Returns
        -------
        node: NodeView
            The node view
        """
        return NodeView(self)

    def get_node_num(self) -> int:
        """
        Get the number of nodes in the graph.

        Returns
        -------
        num_nodes: int
            The number of nodes in the graph.
        """
        return len(self._node_attributes)

    def add_nodes(self, node_num: int) -> None:
        """
        Add a number of nodes to the graph.

        Parameters
        ------
        node_num: int
            The number of nodes to be added
        """
        assert node_num > 0, "The number of nodes to be added should be greater than 0. (Got {})".format(node_num)
        current_num_nodes = self.get_node_num()

        # Create placeholders in the node attribute dictionary
        for new_node_idx in range(current_num_nodes, current_num_nodes + node_num):
            self._node_attributes[new_node_idx] = single_node_attr_factory(**res_init_node_attr)

        # Do padding in the node feature dictionary
        for key in self._node_features.keys():
            self._node_features[key] = entail_zero_padding(self._node_features[key], node_num)

    # Node feature operations
    @property
    def node_features(self) -> NodeFeatView:
        """
        Access and modify node feature vectors (tensor).
        This property can be accessed in a dict-of-dict fashion, with the order being [name][index].
        'name' indicates the name of the feature vector. 'index' selects the specific nodes to be accessed.
        When accessed independently, returns the feature dictionary with the format {name: tensor}

        Examples
        --------
        >>> g = GraphData()
        >>> g.add_nodes(10)
        >>> import torch
        >>> g.node_features['x'] = torch.rand((10, 10))
        >>> g.node_features['x'][0]
        torch.Tensor([0.1036, 0.6757, 0.4702, 0.8938, 0.6337, 0.3290, 0.6739, 0.1091, 0.7996, 0.0586])

        Returns
        -------
        NodeFeatView
        """

        return self.nodes[:].features

    def get_node_features(self, nodes: int or slice) -> torch.Tensor:
        """
        Get the node feature dictionary of the `nodes`

        Parameters
        ----------
        nodes: int or slice
            The nodes to be accessed

        Returns
        -------
        node_features: dict
            The reference dict of the actual tensor
        """
        ret = dict()
        for key in self._node_features.keys():
            if self._node_features[key] is None:
                ret[key] = None
            else:
                ret[key] = self._node_features[key][nodes]
        return ret

    def get_node_feature_names(self):
        return self._node_features.keys()

    def set_node_features(self, nodes: int or slice, new_data: dict) -> None:
        """
        Set the features of the `nodes` with the given `new_data``.

        Parameters
        ----------
        nodes: int or slice
            The nodes involved
        new_data: dict
            The new data to write. Key indicates feature name and value indicates the actual value

        Raises
        ----------
        SizeMismatchException
            If the size of the new features does not match the node number
        """
        # Consistency check
        for key in new_data.keys():
            if key not in self._node_features or self._node_features[key] is None:  # A new feature is added
                # If the shape of the new feature does not match the number of existing nodes, then error occurs
                if (not isinstance(nodes, slice)) or (
                        len(slice_to_list(nodes, self.get_node_num())) != self.get_node_num()):
                    raise ValueError(
                        'The new feature `{}\' should cover all existing {} nodes!'.format(key, self.get_node_num()))

        # Modification
        for key, value in new_data.items():
            # Node-shape check
            assert value.shape[0] == self.get_node_num(), \
                "The shape feature '{}' does not match the number of nodes in the graph. Got a {} tensor but have {} nodes.".format(
                    key, value.shape, self.get_node_num())

            assert isinstance(value, torch.Tensor), "`{}' is not a tensor. Node features are expected to be tensor."
            if key not in self._node_features or self._node_features[key] is None:
                self._node_features[key] = value
            else:
                if nodes == slice(None, None, None):
                    self._node_features[key] = value
                else:
                    self._node_features[key][nodes] = value

    # Node attribute operations
    @property
    def node_attributes(self) -> dict:
        """
        Access node attribute dictionary

        Returns
        -------
        node_attribute_dict: dict
            The dict of node attributes
        """
        return self._node_attributes

    def get_node_attrs(self, nodes: int or slice):
        """
        Get the attributes of the given `nodes`.

        Parameters
        ----------
        nodes: int
         or slice
            The given node index

        Returns
        -------
        dict
            The node attribute dictionary.
        """
        if isinstance(nodes, slice):
            node_idx = slice_to_list(nodes, self.get_node_num())
        else:
            node_idx = [nodes]

        ret = {}
        for idx in node_idx:
            ret[idx] = self._node_attributes[idx]
        return ret

    # Edge views and operations
    @property
    def edges(self):
        """
        Return an edge view of the edges and the corresponding data

        Returns
        -------
        edges: EdgeView
        """
        return EdgeView(self)

    def get_edge_num(self) -> int:
        """
        Get the number of edges in the graph

        Returns
        -------
        num_edges: int
            The number of edges
        """
        return len(self._edge_indices.src)

    def add_edge(self, src: int, tgt: int):
        """
        Add one edge to the graph.

        Parameters
        ----------
        src: int
            Source node index
        tgt: int
            Target node index

        Raises
        ------
        ValueError
            If one of the endpoints of the edge doesn't exist in the graph.
        """
        # Consistency check
        if (src < 0 or src >= self.get_node_num()) and (tgt < 0 and tgt >= self.get_node_num()):
            raise ValueError('Endpoint not in the graph.')

        # Duplicate edge check. If the edge to be added already exists in the graph, then skip it.
        endpoint_tuple = (src, tgt)
        if endpoint_tuple in self._nids_eid_mapping.keys():
            warnings.warn('Edge {} is already in the graph. Skipping this edge.'.format(endpoint_tuple), Warning)
            return

        # Append to the mapping list
        eid = self.get_edge_num()
        self._eid_nids_mapping[eid] = endpoint_tuple
        self._nids_eid_mapping[endpoint_tuple] = eid

        # Add edge
        self._edge_indices.src.append(src)
        self._edge_indices.tgt.append(tgt)

        # Initialize edge feature and attribute
        # 1. create placeholder in edge attribute dictionary
        self._edge_attributes[eid] = single_edge_attr_factory(res_init_edge_attributes)
        # 2. perform zero padding
        for key in self._edge_features.keys():
            self._edge_features[key] = entail_zero_padding(self._edge_features[key], 1)

    def add_edges(self, src: int or list, tgt: int or list):
        """
        Add a bunch of edges to the graph.

        Parameters
        ----------
        src: int or list
            Source node indices
        tgt: int or list
            Target node indices

        Raises
        ------
        ValueError
            If the lengths of `src` and `tgt` don't match or one of the list contains no element.
        """
        src, tgt = check_and_expand(int_to_list(src), int_to_list(tgt))
        for src_idx, tgt_idx in zip(src, tgt):
            self.add_edge(src_idx, tgt_idx)

    def edge_ids(self, src: int or list, tgt: int or list) -> list:
        """
        Convert the given endpoints to edge indices.

        Parameters
        ----------
        src: int or list
            The index of source node(s).
        tgt: int or list
            The index of target node(s).

        Returns
        -------
        list
            The index of corresponding edges.

        Raises
        ------
        EdgeNotFoundException
            If the edge is not in the graph.
        """
        assert isinstance(src, int) or isinstance(src, list), "`src` should be either int or list."
        assert isinstance(tgt, int) or isinstance(tgt, list), "`tgt` should be either int or list."
        src, tgt = check_and_expand(int_to_list(src), int_to_list(tgt))
        eid_list = []
        try:
            for src_idx, tgt_idx in zip(src, tgt):
                eid_list.append(self._nids_eid_mapping[(src_idx, tgt_idx)])
        except KeyError:
            raise EdgeNotFoundException('Edge {} does not exist!'.format((src, tgt)))
        return eid_list

    def get_all_edges(self) -> list:
        """
        Get all the edges in the graph

        Returns
        -------
        edges: list
            List of edges. Each edge is in the shape of the endpoint tuple (src, dst).
        """
        edges = []
        for i in range(self.get_edge_num()):
            edges.append((self._edge_indices.src[i], self._edge_indices.tgt[i]))
        return edges

    # Edge feature operations
    @property
    def edge_features(self):
        return self.edges[:].features

    def get_edge_feature(self, edges: list):
        """
        Get the feature of the given edges.

        Parameters
        ----------
        edges: list
            Edge indices

        Returns
        -------
        dict
            The dictionary containing all relevant features.
        """
        ret = {}
        for key in self._edge_features.keys():
            if self._edge_features[key] is None:
                ret[key] = None
            else:
                ret[key] = self._edge_features[key][edges]
        return ret

    def get_edge_feature_names(self):
        """Get all the names of edge features"""
        return self._edge_features.keys()

    def set_edge_feature(self, edges: int or slice or list, new_data: dict):
        """
        Set edge feature

        Parameters
        ----------
        edges: list
            Edge indices
        new_data: dict
            New data

        Raises
        ----------
        SizeMismatchException
            If the size of the new features does not match the node number
        """
        # Consistency check
        for key in new_data.keys():
            if key not in self._edge_features or self._edge_features[key] is None:  # A new feature is added
                # If the shape of the new feature does not match the number of existing nodes, then error occurs
                if (not isinstance(edges, slice)) or (
                        len(slice_to_list(edges, self.get_edge_num())) != self.get_edge_num()):
                    raise SizeMismatchException(
                        'The new feature `{}\' should cover all existing {} edges!'.format(key, self.get_edge_num()))

        # Modification
        for key, value in new_data.items():
            assert isinstance(value, torch.Tensor), "`{}' is not a tensor. Node features are expected to be tensor."

            assert value.shape[0] == self.get_edge_num(), "Length of the feature vector does not match the edge number." \
                                                          "Got tensor '{}' of shape {} but the graph has only {} edges.".format(
                key, value.shape, self.get_edge_num())

            if key not in self._edge_features or self._edge_features[key] is None:
                self._edge_features[key] = value
            elif edges == slice(None, None, None):
                # Same as node features, if the edges to be modified is all the edges in the graph.
                self._edge_features[key] = value
            else:
                self._edge_features[key][edges] = value

    # Edge attribute operations
    @property
    def edge_attributes(self):
        return self._edge_attributes

    # Conversion utility functions
    def to_dgl(self) -> dgl.DGLGraph:
        """
        Convert to dgl.DGLGraph

        Returns
        -------
        g: dgl.DGLGraph
            The converted dgl.DGLGraph
        """
        dgl_g = dgl.DGLGraph().to(self.device)
        # Add nodes and their features
        dgl_g.add_nodes(num=self.get_node_num())
        for key, value in self._node_features.items():
            if value is not None:
                dgl_g.ndata[key] = value
        # Add edges and their features
        dgl_g.add_edges(u=self._edge_indices.src, v=self._edge_indices.tgt)
        for key, value in self._edge_features.items():
            if value is not None:
                dgl_g.edata[key] = value
        return dgl_g

    def from_dgl(self, dgl_g: dgl.DGLGraph):
        """
        Build the graph from dgl.DGLGraph

        Parameters
        ----------
        dgl_g: dgl.DGLGraph
            The source graph
        """
        assert self.get_edge_num() == 0 and self.get_node_num() == 0, \
            'This graph isn\'t an empty graph. Please use an empty graph for conversion.'

        # Add nodes
        self.add_nodes(dgl_g.number_of_nodes())
        for k, v in dgl_g.ndata.items():
            self.node_features[k] = v

        # Add edges
        src_tensor, tgt_tensor = dgl_g.edges()
        src_list = list(src_tensor.detach().cpu().numpy())
        tgt_list = list(tgt_tensor.detach().cpu().numpy())
        self.add_edges(src_list, tgt_list)
        for k, v in dgl_g.edata.items():
            self.edge_features[k] = v

    def from_dense_adj(self, adj: torch.Tensor):
        assert adj.dim() == 2, 'Adjancency matrix is not 2-dimensional.'
        assert adj.shape[0] == adj.shape[1], 'Adjancecy is not a square.'

        node_num = adj.shape[0]
        self.add_nodes(node_num)
        edge_weight = []
        for i in range(adj.shape[0]):
            for j in range(adj.shape[1]):
                if adj[i][j] != 0:
                    self.add_edge(i, j)
                    edge_weight.append(adj[i][j])
        edge_weight = torch.stack(edge_weight, dim=0)
        self.edge_features['edge_weight'] = edge_weight

    def from_scipy_sparse_matrix(self, adj: scipy.sparse.coo_matrix):
        assert adj.shape[0] == adj.shape[1], 'Got an adjancecy matrix which is not a square.'

        num_nodes = adj.shape[0]
        self.add_nodes(num_nodes)

        for i in range(adj.row.shape[0]):
            self.add_edge(adj.row[i], adj.col[i])
        self.edge_features['edge_weight'] = torch.tensor(adj.data)

    def adj_matrix(self):
        ret = torch.zeros((self.get_node_num(), self.get_node_num()))
        all_edges = self.edges()
        for i in range(len(all_edges)):
            u, v = all_edges[i]
            ret[u][v] = 1
        return ret

    def scipy_sparse_adj(self):
        row = np.array(self._edge_indices[0])
        col = np.array(self._edge_indices[1])
        data = np.ones(self.get_edge_num())
        matrix = scipy.sparse.coo_matrix((data, (row, col)), shape=(self.get_node_num(), self.get_node_num()))
        return matrix

    def from_graphdata(self, src):
        """Build a clone from a source GraphData"""
        self.add_nodes(src.get_node_num())
        for src_node, tgt_node in src.get_all_edges():
            self.add_edge(src_node, tgt_node)
        for i in range(src.get_node_num()):
            self._node_attributes[i] = src.node_attributes[i]
        for feat_name in src.get_node_feature_names():
            self._node_features[feat_name] = src.node_features[feat_name]
        for i in range(src.get_edge_num()):
            self._edge_attributes[i] = src.edge_attributes[i]
        for feat_name in src.get_edge_feature_names():
            self._edge_features[feat_name] = src.edge_features[feat_name]
        for k, v in src.graph_attributes.items():
            self.graph_attributes[k] = v

    def union(self, graph):
        """
        Merge a graph into current graph.

        Parameters
        ----------
        graph: GraphData
            The new graph to be merged.
        """

        # Consistency check
        # 0. check if both graphs have nodes
        assert self.get_node_num() > 0 and graph.get_node_num() > 0, \
            "Both participants of the union operation should contain at least 1 node."

        # 1. check if node feature names are consistent
        for feat_name in graph.node_features.keys():
            assert feat_name in self._node_features.keys(), "Node feature '{}' does not exist in current graph.".format(
                feat_name)

        # 2. check if edge feature names are consistent
        for feat_name in graph.edge_features.keys():
            assert feat_name in self._edge_features.keys(), "Edge feature '{}' does not exist in current graph.".format(
                feat_name)

        # Save original information
        old_node_num = self.get_node_num()
        old_edge_num = self.get_edge_num()

        # Append information to current graph
        # 1. add nodes
        self.add_nodes(graph.get_node_num())

        # 2. add node features
        append_node_st_idx = old_node_num
        append_node_ed_idx = self.get_node_num()
        for feat_name in graph.node_features.keys():
            if self.node_features[feat_name] is None:
                continue
            self.node_features[feat_name][append_node_st_idx:append_node_ed_idx] = graph.node_features[feat_name]

        # 3. add node attributes
        for node_idx in range(append_node_st_idx, append_node_ed_idx):
            self.node_attributes[node_idx] = graph.node_attributes[node_idx - append_node_st_idx]

        # 4. add edges
        for (src, tgt) in graph.edges():
            self.add_edge(src + old_node_num, tgt + old_node_num)

        # 5. add edge features
        append_edge_st_idx = old_edge_num
        append_edge_ed_idx = self.get_edge_num()
        for feat_name in graph.edge_features.keys():
            if self.edge_features[feat_name] is None:
                continue
            self.edge_features[feat_name][append_edge_st_idx:append_edge_ed_idx] = graph.edge_features[feat_name]

        # 6. add edge attributes
        for edge_idx in range(append_edge_st_idx, append_edge_ed_idx):
            self.edge_attributes[edge_idx] = graph.edge_attributes[edge_idx - append_edge_st_idx]

    def split(self, node_st_idx: int, node_ed_idx: int):
        """
        Given the starting and ending indices of the nodes, split it out of the large graph.

        The corresponding subgraph indicated by the node indices should be a connected component of the large graph and
        should have no connection to other nodes in the graph. Otherwise it cannot be split from the large graph without
        information loss.

        The node indices indicate by `node_st_idx` and `node_ed_idx` is the closed set [`node_st_idx`, `node_ed_idx`].

        Parameters
        ----------
        node_st_idx: int
            The starting node index of the subgraph in the large graph.
        node_ed_idx: int
            The ending node index of the subgraph in the large graph.

        Returns
        -------
        GraphData
            The extracted subgraph.

        Raises
        ------
        ValueError
            If the subgraph has connection with other nodes in the large graph.
        """
        assert node_ed_idx >= node_st_idx, "Got node_ed_idx({}) > node_st_idx({}). " \
                                           "The subgraph should contain at least 1 node.".format(node_ed_idx,
                                                                                                 node_st_idx)
        # extract the corresponding edges from the large graph
        all_edges = self.get_all_edges()
        node_idx_range = range(node_st_idx, node_ed_idx + 1)
        subgraph_edges = []
        for i in range(len(all_edges)):
            current_edge: (int, int) = all_edges[i]
            src, tgt = current_edge
            if src not in node_idx_range and tgt not in node_idx_range:
                continue
            elif src in node_idx_range and tgt in node_idx_range:
                subgraph_edges.append(current_edge)
            else:
                raise ValueError("The subgraph to be extracted has connection with other nodes in the large graph.")
        # convert the edge from node index tuples to edge indices
        no_edge = False
        if len(subgraph_edges) != 0:
            subgraph_edge_src = [src for (src, tgt) in subgraph_edges]
            subgraph_edge_tgt = [tgt for (src, tgt) in subgraph_edges]
            subgraph_edge_ids = self.edge_ids(subgraph_edge_src, subgraph_edge_tgt)
            subgraph_edge_st_idx = min(subgraph_edge_ids)
            subgraph_edge_ed_idx = max(subgraph_edge_ids)
        else:
            no_edge = True
        # build the subgraph
        subgraph = GraphData()
        subgraph.add_nodes(node_ed_idx - node_st_idx + 1)
        for k, v in self._node_features.items():
            if v is not None:
                subgraph.node_features[k] = v[node_st_idx:node_ed_idx + 1]
        for i in range(node_st_idx, node_ed_idx + 1, 1):
            subgraph.node_attributes[i - node_st_idx] = self._node_attributes[i]

        if not no_edge:
            for src, tgt in subgraph_edges:
                subgraph.add_edge(src - node_st_idx, tgt - node_st_idx)
            for k, v in self._edge_features.items():
                if v is not None:
                    subgraph.edge_features[k] = v[subgraph_edge_st_idx:subgraph_edge_ed_idx + 1]
            for i in range(subgraph_edge_st_idx, subgraph_edge_ed_idx + 1, 1):
                subgraph.edge_attributes[i - subgraph_edge_st_idx] = self._edge_attributes[i]
        return subgraph

    def copy_batch_info(self, batch):
        self.batch = batch.batch
        self.device = batch.device
        self.batch_size = batch.batch_size
        self._batch_num_edges = batch._batch_num_edges
        self._batch_num_nodes = batch._batch_num_nodes


def from_dgl(g: dgl.DGLGraph) -> GraphData:
    """
    Convert a dgl.DGLGraph to a GraphData object.

    Parameters
    ----------
    g: dgl.DGLGraph
        The source graph in DGLGraph format.

    Returns
    -------
    GraphData
        The converted graph in GraphData format.
    """
    graph = GraphData()
    graph.from_dgl(g)
    return graph


def to_batch(graphs: list = None) -> GraphData:
    """
    Convert a list of GraphData to a large graph (a batch).

    Parameters
    ----------
    graphs: list of GraphData
        The list of GraphData to be batched

    Returns
    -------
    GraphData
        The large graph containing all the graphs in the batch.
    """
    batch = GraphData(graphs[0], graphs[0].device)
    batch.batch_size = len(graphs)
    batch.batch = [0] * graphs[0].get_node_num()
    batch._batch_num_nodes = [g.get_node_num() for g in graphs]
    batch._batch_num_edges = [g.get_edge_num() for g in graphs]
    for i in range(1, len(graphs)):
        batch.union(graphs[i])
        batch.batch += [i] * graphs[i].get_node_num()
    return batch


def from_batch(batch: GraphData) -> list:
    """
    Convert a batch consisting of several GraphData instances to a list of GraphData instances.

    Parameters
    ----------
    batch: GraphData
        The source batch to be split.

    Returns
    -------
    list
        A list containing all the GraphData instances contained in the source batch.
    """

    num_nodes = batch._batch_num_nodes
    num_edges = batch._batch_num_edges
    all_edges = batch.get_all_edges()
    batch_size = batch.batch_size
    ret = []
    cum_n_nodes = 0
    cum_n_edges = 0

    # Construct graph respectively
    for i in range(batch_size):
        g = GraphData(device=batch.device)
        g.add_nodes(num_nodes[i])
        edges = all_edges[cum_n_edges:cum_n_edges + num_edges[i]]
        src, tgt = [e[0] - cum_n_nodes for e in edges], [e[1] - cum_n_nodes for e in edges]
        g.add_edges(src, tgt)
        cum_n_edges += num_edges[i]
        cum_n_nodes += num_nodes[i]
        ret.append(g)

    # Add node and edge features
    for k, v in batch._node_features.items():
        if v is not None:
            cum_n_nodes = 0  # Cumulative node numbers
            for i in range(batch_size):
                ret[i].node_features[k] = v[cum_n_nodes:cum_n_nodes + num_nodes[i]]
                cum_n_nodes += num_nodes[i]

    for k, v in batch._edge_features.items():
        if v is not None:
            cum_n_edges = 0  # Cumulative edge numbers
            for i in range(batch_size):
                ret[i].edge_features[k] = v[cum_n_edges:cum_n_edges + num_edges[i]]
                cum_n_edges += num_edges[i]
    cum_n_nodes = 0
    cum_n_edges = 0

    # Add node and edge attributes
    for graph_cnt in range(batch_size):
        for num_graph_nodes in range(num_nodes[graph_cnt]):
            ret[graph_cnt].node_attributes[num_graph_nodes] = batch.node_attributes[cum_n_nodes + num_graph_nodes]
        for num_graph_edges in range(num_edges[graph_cnt]):
            ret[graph_cnt].edge_attributes[num_graph_edges] = batch.edge_attributes[cum_n_edges + num_graph_edges]
        cum_n_edges += num_edges[graph_cnt]
        cum_n_nodes += num_nodes[graph_cnt]

    return ret


# Testing code
if __name__ == '__main__':
    a1 = GraphData()
    a1.add_nodes(10)
    a2 = GraphData()
    a2.add_nodes(15)
    a3 = to_batch([a1, a2])
    print(len(a1.node_attributes), len(a2.node_attributes), len(a3.node_attributes))



import os
import json
import random

random.seed(10)

def process_stories(root_dir):
    stories = os.listdir(root_dir)
    # random.shuffle(stories)
    # stories = stories[:50000]
    examples = []
    cnt_words = []
    for i, f_name in enumerate(stories):
        if i%100==0:
            print(i)
        with open(root_dir+f_name, 'r') as f:
            text = f.readlines()
            # examples.append(text)
            cnt_words.append(len(' '.join(text).split(' ')))
            example_dict = {'article':[],
                            'highlight':[]}
            is_hightlight = False
            for line in text:
                if line == '\n':
                    continue
                if line == '@highlight\n':
                    is_hightlight = True
                    continue
                if is_hightlight:
                    example_dict['highlight'].append([line.strip()])
                    is_hightlight = False
                else:
                    example_dict['article'].append(line.strip())

            examples.append(example_dict)

    # print(sum(cnt_words)/len(cnt_words))
    return examples

example1 = process_stories('raw/cnn_stories_tokenized/')
# example2 = process_stories('raw/dm_stories_tokenized/')

# examples = example1 + example2
examples = example1
random.shuffle(examples)

# train_30 = examples[:30]
# train_3k = examples[:3000]
# train_1w = examples[:10000]
train_3w = examples[:30000]
# train_9w = examples[:90000]

# with open('raw/train_30.json', 'w+') as f:
#     json.dump(train_30, f, indent=1)

# with open('raw/train_3k.json', 'w+') as f:
#     json.dump(train_3k, f, indent=1)

# with open('raw/train_1w.json', 'w+') as f:
#     json.dump(train_1w, f, indent=1)

with open('raw/train_3w.json', 'w+') as f:
    json.dump(train_3w, f, indent=1)

# with open('raw/train_9w.json', 'w+') as f:
#     json.dump(train_9w, f, indent=1)

# with open('raw/val.json', 'w+') as f:
#     json.dump(examples[-6000:-3000], f, indent=1)

# with open('raw/test.json', 'w+') as f:
#     json.dump(examples[-3000:], f, indent=1)

# print('num_skip_long_input='+str(num_skip_long_input))
a = 0


import os

os.environ["CUDA_VISIBLE_DEVICES"] = "2"
# os.environ['CUDA_LAUNCH_BLOCKING'] = "5"

from graph4nlp.pytorch.data.data import GraphData
from graph4nlp.pytorch.datasets.kinship import KinshipDataset
from graph4nlp.pytorch.modules.graph_embedding.gat import GAT, GATLayer
from graph4nlp.pytorch.modules.graph_embedding.graphsage import GraphSAGELayer
from graph4nlp.pytorch.modules.graph_embedding.ggnn import GGNN, GGNNLayer
from graph4nlp.pytorch.modules.graph_embedding.gcn import GCNLayer
from graph4nlp.pytorch.modules.utils.vocab_utils import Vocab
from graph4nlp.pytorch.modules.prediction.classification.kg_completion.DistMult import DistMult
from graph4nlp.pytorch.modules.utils.generic_utils import to_cuda
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
import torch.optim as optim
import dgl
import numpy as np
import copy
import random

from graph4nlp.pytorch.modules.evaluation.base import EvaluationMetricBase
from graph4nlp.pytorch.modules.graph_construction.embedding_construction import EmbeddingConstruction
from graph4nlp.pytorch.modules.loss.kg_loss import *


class RankingAndHits(EvaluationMetricBase):
    def __init__(self, model_path='best_graph2distmult', batch_size=64):
        super(RankingAndHits, self).__init__()
        self.batch_size = batch_size
        self.best_mrr = 0.
        self.best_hits1 = 0.
        self.best_hits10 = 0.
        self.model_path = model_path

    def calculate_scores(self, model, dataloader, name, kg_graph, device, save=True):
        print('')
        print('-' * 50)
        print(name)
        hits_left = []
        hits_right = []
        hits = []
        ranks = []
        ranks_left = []
        ranks_right = []
        for i in range(10):
            hits_left.append([])
            hits_right.append([])
            hits.append([])
        with open('output_model.txt', 'w') as file:
            for data in dataloader:
                e1, rel, e2_multi, e2_multi_tensor_idx, \
                e2, rel_eval, e1_multi, e1_multi_tensor_idx = data['e1'], data['rel'], \
                                                              data['e2_multi'], data['e2_multi_tensor_idx'], \
                                                              data['e2'], data['rel_eval'], \
                                                              data['e1_multi'], data['e1_multi_tensor_idx']

                e1 = e1.view(-1, 1).to(device)
                rel = rel.view(-1, 1).to(device)
                e2_multi = e2_multi.to(device)
                e2_multi_tensor_idx = e2_multi_tensor_idx.to(device)

                e2 = e2.view(-1, 1).to(device)
                rel_eval = rel_eval.view(-1, 1).to(device)
                e1_multi = e1_multi.to(device)
                e1_multi_tensor_idx = e1_multi_tensor_idx.to(device)
                pred1 = model(kg_graph, e1, rel, e2_multi, require_loss=False)
                pred2 = model(kg_graph, e2, rel_eval, e1_multi, require_loss=False)
                pred1, pred2 = pred1.data, pred2.data
                e1, e2 = e1.data, e2.data
                e2_multi1, e2_multi2 = e2_multi_tensor_idx.data, e1_multi_tensor_idx.data

                if e1.size()[0] != self.batch_size:
                    continue

                for i in range(self.batch_size):
                    # these filters contain ALL labels
                    filter1 = e2_multi1[i].long()
                    filter2 = e2_multi2[i].long()

                    num = e1[i, 0].item()
                    # save the prediction that is relevant
                    target_value1 = pred1[i, e2.cpu().numpy()[i, 0].item()].item()
                    target_value2 = pred2[i, e1.cpu().numpy()[i, 0].item()].item()
                    # zero all known cases (this are not interesting)
                    # this corresponds to the filtered setting
                    pred1[i][filter1] = 0.0
                    pred2[i][filter2] = 0.0
                    # write base the saved values
                    pred1[i][e2[i]] = target_value1
                    pred2[i][e1[i]] = target_value2

                    # print(e1[i, 0])

                # sort and rank
                max_values1, argsort1 = torch.sort(pred1, 1, descending=True)
                max_values2, argsort2 = torch.sort(pred2, 1, descending=True)

                argsort1 = argsort1.cpu().numpy()
                argsort2 = argsort2.cpu().numpy()
                for i in range(self.batch_size):
                    # find the rank of the target entities
                    rank1 = np.where(argsort1[i] == e2.cpu().numpy()[i, 0])[0][0]
                    if model.loss_name in ['SoftplusLoss', 'SigmoidLoss'] and max_values1[i][rank1] == max_values1[i][
                        0]:
                        rank1 = 0
                    rank2 = np.where(argsort2[i] == e1.cpu().numpy()[i, 0])[0][0]
                    if model.loss_name in ['SoftplusLoss', 'SigmoidLoss'] and max_values2[i][rank2] == max_values2[i][
                        0]:
                        rank2 = 0
                    # rank+1, since the lowest rank is rank 1 not rank 0
                    ranks.append(rank1 + 1)
                    ranks_left.append(rank1 + 1)
                    ranks.append(rank2 + 1)
                    ranks_right.append(rank2 + 1)

                    file.write(str(e2.cpu().numpy()[i, 0].item()) + '\t')
                    file.write(str(rank1 + 1) + '\n')
                    file.write(str(e1.cpu().numpy()[i, 0].item()) + '\t')
                    file.write(str(rank2 + 1) + '\n')

                    # this could be done more elegantly, but here you go
                    for hits_level in range(10):
                        if rank1 <= hits_level:
                            hits[hits_level].append(1.0)
                            hits_left[hits_level].append(1.0)
                        else:
                            hits[hits_level].append(0.0)
                            hits_left[hits_level].append(0.0)

                        if rank2 <= hits_level:
                            hits[hits_level].append(1.0)
                            hits_right[hits_level].append(1.0)
                        else:
                            hits[hits_level].append(0.0)
                            hits_right[hits_level].append(0.0)

        for i in range(10):
            print('Hits left @{0}: {1}'.format(i + 1, np.mean(hits_left[i])))
            print('Hits right @{0}: {1}'.format(i + 1, np.mean(hits_right[i])))
            print('Hits @{0}: {1}'.format(i + 1, np.mean(hits[i])))
        print('Mean rank left: ', np.mean(ranks_left))
        print('Mean rank right: ', np.mean(ranks_right))
        print('Mean rank: ', np.mean(ranks))
        print('Mean reciprocal rank left: ', np.mean(1. / np.array(ranks_left)))
        print('Mean reciprocal rank right: ', np.mean(1. / np.array(ranks_right)))
        print('Mean reciprocal rank: ', np.mean(1. / np.array(ranks)))

        if np.mean(hits[0]) > self.best_hits1:
            self.best_hits1 = np.mean(hits[0])

        if np.mean(hits[9]) > self.best_hits10:
            self.best_hits10 = np.mean(hits[9])

        if np.mean(1. / np.array(ranks)) > self.best_mrr:
            self.best_mrr = np.mean(1. / np.array(ranks))
            if save:
                print('saving best model...')
                torch.save(model.state_dict(), self.model_path)

        print('-' * 50)
        print('')

class Graph2DistMult(nn.Module):
    def __init__(self, args, vocab, device, num_entities, hidden_size=300, num_relations=None, direction_option='uni',
                 loss_name='BCELoss'):
        super(Graph2DistMult, self).__init__()
        self.args = args
        self.device = device
        self.vocab = vocab
        self.num_entities = num_entities

        self.node_emb = nn.Embedding(num_entities, hidden_size)

        embedding_style = {'single_token_item': False,
                           'emb_strategy': 'w2v_bilstm',
                           'num_rnn_layers': 1}

        self.embedding_layer = EmbeddingConstruction(self.vocab.in_word_vocab,
                                                     embedding_style['single_token_item'],
                                                     embedding_style['emb_strategy'],
                                                     hidden_size=hidden_size,
                                                     fix_word_emb=False,
                                                     fix_bert_emb=False,
                                                     device=self.device)
        self.direction_option = direction_option
        self.hidden_size = hidden_size

        if args.gnn == 'ggnn':
            self.num_layers = 2  # ggnn
            self.gnn_encoder = nn.ModuleList(
                [GGNNLayer(hidden_size, hidden_size, direction_option, num_layers=1, n_etypes=1, bias=True)
                 for i in range(self.num_layers)])
        elif args.gnn == 'gat':
            self.num_layers = 1  # gat uni/bi_fuse/bi_sep
            self.gnn_encoder = nn.ModuleList(
                [GATLayer(hidden_size, hidden_size, num_heads=1, direction_option=direction_option)
                 for i in range(self.num_layers)])
        elif args.gnn == 'graphsage':
            self.num_layers = 2  # graphsage
            self.gnn_encoder = nn.ModuleList(
                [GraphSAGELayer(hidden_size, hidden_size, aggregator_type='mean',
                                direction_option=direction_option)
                 for i in range(self.num_layers)])
        elif args.gnn == 'gcn':
            self.num_layers = 1  # gcn
            self.gnn_encoder = nn.ModuleList(
                [GCNLayer(hidden_size, hidden_size, direction_option=direction_option)
                 for i in range(self.num_layers)])

        self.bn_list = nn.ModuleList([torch.nn.BatchNorm1d(hidden_size)
                                      for i in range(self.num_layers)])  # necessary for this task

        self.distmult = DistMult(rel_emb_from_gnn=False, num_relations=num_relations,
                                 embedding_dim=hidden_size, loss_name=loss_name)

        self.loss_name = loss_name
        if loss_name == 'BCELoss':
            self.loss = torch.nn.BCELoss()
        elif loss_name == "SoftplusLoss":
            self.loss = SoftplusLoss()
        elif loss_name == "SigmoidLoss":
            self.loss = SigmoidLoss()
        elif loss_name == "SoftMarginLoss":
            self.loss = nn.SoftMarginLoss()
        elif loss_name == "MSELoss":
            self.loss = nn.MSELoss()
        else:
            raise NotImplementedError()

        self.reset_parameters()

    def embedding(self, kg_graph):
        graph_nodes_idx = []
        node_len_list = []
        for x in kg_graph.graph_attributes['graph_nodes']:
            tmp_list = []
            for token in x.split(' '):
                tmp_list.append(self.vocab.in_word_vocab.word2index[token])

            graph_nodes_idx.append(tmp_list)
            node_len_list.append(len(tmp_list))

        max_size = max(node_len_list)
        graph_nodes_idx = [x + [self.vocab.in_word_vocab.PAD] * (max_size - len(x)) for x in graph_nodes_idx]
        graph_nodes_idx = torch.tensor(graph_nodes_idx, dtype=torch.long).to(self.device)
        node_len_tensor = torch.tensor(node_len_list, dtype=torch.long).to(self.device)
        num_nodes = torch.tensor([len(node_len_list)], dtype=torch.long).to(self.device)
        kg_graph.node_features['token_id'] = graph_nodes_idx
        node_feat = self.embedding_layer(kg_graph, node_len_tensor, num_nodes)
        return node_feat

    def reset_parameters(self):
        nn.init.xavier_normal_(self.node_emb.weight.data)

    def forward(self, kg_graph, e1, rel, e2_multi=None, require_loss=True):
        kg_graph = to_cuda(kg_graph, self.device)
        kg_graph.node_features['node_feat'] = self.embedding(kg_graph)
        list_e_r_pair_idx = list(zip(e1.squeeze().tolist(), rel.squeeze().tolist()))

        # run GNN
        # ================================================================================
        node_feats = kg_graph.node_features['node_feat']
        dgl_graph = kg_graph.to_dgl()

        if self.direction_option == 'undirected':
            node_embs = kg_graph.node_features['node_feat']
            for i in range(self.num_layers):
                if self.args.gnn == 'ggnn' or self.args.gnn == 'gcn' or self.args.gnn == 'graphsage':
                    node_embs = self.gnn_encoder[i](dgl_graph, node_embs)  # GGNN
                elif self.args.gnn == 'gat':
                    node_embs = self.gnn_encoder[i](dgl_graph, node_embs).squeeze()  # GAT
                node_embs = torch.dropout(torch.tanh(self.bn_list[i](node_embs)), 0.25, train=require_loss)
        else:
            assert node_feats.shape[1] == self.hidden_size

            zero_pad = node_feats.new_zeros((node_feats.shape[0], self.hidden_size - node_feats.shape[1]))
            node_feats = torch.cat([node_feats, zero_pad], -1)

            feat_in = node_feats
            feat_out = node_feats

            if self.direction_option == 'bi_sep':
                for i in range(self.num_layers):
                    h = self.gnn_encoder[i](dgl_graph, (feat_in, feat_out))
                    if self.args.gnn == 'ggnn' or self.args.gnn == 'graphsage' or self.args.gnn == 'gcn':
                        feat_in = torch.dropout(torch.tanh(self.bn_list[i](h[0])), 0.25, train=require_loss)
                        feat_out = torch.dropout(torch.tanh(self.bn_list[i](h[1])), 0.25, train=require_loss)
                    elif self.args.gnn == 'gat':
                        feat_in = torch.dropout(torch.tanh(self.bn_list[i](h[0].squeeze())), 0.25,
                                                train=require_loss)  # GAT
                        feat_out = torch.dropout(torch.tanh(self.bn_list[i](h[1].squeeze())), 0.25,
                                                 train=require_loss)  # GAT
                node_embs = (feat_in + feat_out) / 2
            elif self.direction_option == 'bi_fuse':
                for i in range(self.num_layers):
                    if self.args.gnn == 'ggnn':
                        h = self.gnn_encoder[i](dgl_graph, (feat_in, feat_in))  # GGNN
                        feat_in = torch.dropout(torch.tanh(self.bn_list[i](h[0])), 0.25, train=require_loss)  # GGNN
                    elif self.args.gnn == 'gat':
                        h = self.gnn_encoder[i](dgl_graph, feat_in).squeeze()  # GAT
                        feat_in = torch.dropout(torch.tanh(self.bn_list[i](h)), 0.25, train=require_loss)  # GAT
                    elif self.args.gnn == 'graphsage' or self.args.gnn == 'gcn':
                        h = self.gnn_encoder[i](dgl_graph, feat_in)  # GraphSage
                        feat_in = torch.dropout(torch.tanh(self.bn_list[i](h)), 0.25, train=require_loss)  # GraphSage
                node_embs = feat_in
            else:
                raise RuntimeError('Unknown `bidirection` value: {}'.format(self.direction_option))

        kg_graph.node_features['node_emb'] = node_embs
        # ================================================================================

        kg_graph.graph_attributes['list_e_r_pair_idx'] = list_e_r_pair_idx
        kg_graph.graph_attributes['multi_binary_label'] = e2_multi

        # down-task
        kg_graph = self.distmult(kg_graph)
        if require_loss:
            if self.loss_name == "SoftplusLoss" or self.loss_name == "SigmoidLoss":
                loss = self.loss(kg_graph.graph_attributes['p_score'],
                                 kg_graph.graph_attributes['n_score'])
            else:
                loss = self.loss(kg_graph.graph_attributes['logits'], e2_multi)
            return kg_graph.graph_attributes['logits'], loss
        else:
            return kg_graph.graph_attributes['logits']


class Kinship:
    def __init__(self, device, args):
        super(Kinship, self).__init__()
        self.device = device
        self._build_dataloader()
        self._build_model(args)
        self._build_optimizer()
        self._build_evaluation()

    def _build_dataloader(self):
        dataset = KinshipDataset(root_dir='examples/pytorch/kg_completion/kinship',
                                 topology_builder=None,
                                 topology_subdir='e1rel_to_e2')

        # TODO: When edge2node is True, num_entities != train_set.KG_graph.get_node_num()
        self.num_entities = dataset.KG_graph.graph_attributes['num_entities']
        self.num_relations = dataset.KG_graph.graph_attributes['num_relations']
        self.kg_graph = dataset.KG_graph

        self.train_dataloader = DataLoader(dataset.train, batch_size=64, shuffle=True,
                                           num_workers=1,
                                           collate_fn=dataset.collate_fn)

        self.val_dataloader = DataLoader(dataset.val, batch_size=64, shuffle=True,
                                         num_workers=1,
                                         collate_fn=dataset.collate_fn)

        self.test_dataloader = DataLoader(dataset.test, batch_size=64, shuffle=True,
                                          num_workers=1,
                                          collate_fn=dataset.collate_fn)
        self.vocab = dataset.vocab_model

    def _build_model(self, args):  # BCELoss SigmoidLoss
        self.model = Graph2DistMult(args,
                                    self.vocab,
                                    self.device,
                                    num_entities=self.num_entities,
                                    num_relations=self.num_relations,
                                    loss_name=args.loss,
                                    direction_option=args.direction_option).to(self.device)

    def _build_optimizer(self):
        parameters = [p for p in self.model.parameters() if p.requires_grad]
        self.optimizer = optim.Adam(parameters, lr=2e-3)

    def _build_evaluation(self):
        self.metrics = [RankingAndHits(
            model_path='best_graph2distmult_' + self.model.direction_option + '_' + self.model.loss_name)]

    def train(self):
        for epoch in range(30):
            self.model.train()
            loss_list = []

            for data in self.train_dataloader:
                e1, rel, e2_multi, e2_multi_tensor_idx, \
                e2, rel_eval, e1_multi, e1_multi_tensor_idx = data['e1'], data['rel'],\
                                                              data['e2_multi'], data['e2_multi_tensor_idx'], \
                                                              data['e2'], data['rel_eval'], \
                                                              data['e1_multi'], data['e1_multi_tensor_idx']

                e1 = torch.cat([e1, e2]).to(self.device)
                rel = torch.cat([rel, rel_eval]).to(self.device)
                e2_multi = torch.cat([e2_multi, e1_multi], dim=0).to(self.device)

                if random.randint(0, 1) == 0:
                    e1 = torch.cat([e1]).to(self.device)
                    rel = torch.cat([rel]).to(self.device)
                    e2_multi = torch.cat([e2_multi], dim=0).to(self.device)
                else:
                    e1 = torch.cat([e2]).to(self.device)
                    rel = torch.cat([rel_eval]).to(self.device)
                    e2_multi = torch.cat([e1_multi], dim=0).to(self.device)

                _, loss = self.model(self.kg_graph, e1, rel, e2_multi, require_loss=True)
                loss_list.append(loss.item())
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()

            print('#' * 50)
            print("Epoch: {}".format(epoch))
            print('train loss =' + str(sum(loss_list) / len(loss_list)))
            print('#' * 50)
            self.valid()

        return self.metrics[0].best_mrr

    def valid(self):
        self.model.eval()

        self.metrics[0].calculate_scores(self.model,
                                         self.val_dataloader,
                                         name='validation',
                                         kg_graph=self.kg_graph,
                                         device=self.device)

        return self.metrics[0].best_mrr

    def test(self):
        print('load best model...')
        self.model.load_state_dict(torch.load(self.metrics[0].model_path))
        self.model.eval()
        self.metrics[0].calculate_scores(self.model,
                                         self.test_dataloader,
                                         name='test',
                                         kg_graph=self.kg_graph,
                                         device=self.device,
                                         save=False)

        return self.metrics[0].best_mrr


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--gnn', type=str, default='gcn')
    parser.add_argument('--direction_option', type=str, default='undirected')
    parser.add_argument('--loss', type=str, default='SigmoidLoss')
    args = parser.parse_args()

    if torch.cuda.is_available():
        print('using cuda...')
        device = torch.device('cuda:0')
    else:
        device = torch.device('cpu')

    runner = Kinship(device, args)
    max_score = runner.train()
    max_score = runner.test()
    print("Train finish, best MRR: {:.3f}".format(max_score))



import os
import random
import time
import pickle
import argparse
import copy

import dgl
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init
import torch.optim as optim
from stanfordcorenlp import StanfordCoreNLP

from graph4nlp.pytorch.data.data import GraphData, from_batch

from graph4nlp.pytorch.datasets.geo import GeoDatasetForTree

from graph4nlp.pytorch.modules.evaluation.base import EvaluationMetricBase
from graph4nlp.pytorch.modules.graph_construction.dependency_graph_construction import DependencyBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_construction.constituency_graph_construction import ConstituencyBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_construction.node_embedding_based_graph_construction import NodeEmbeddingBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_construction.node_embedding_based_refined_graph_construction import NodeEmbeddingBasedRefinedGraphConstruction

from graph4nlp.pytorch.modules.graph_embedding.gat import GAT
from graph4nlp.pytorch.modules.graph_embedding.ggnn import GGNN
from graph4nlp.pytorch.modules.graph_embedding.graphsage import GraphSAGE
from graph4nlp.pytorch.modules.graph_embedding.gcn import GCN

from graph4nlp.pytorch.modules.prediction.generation.TreeBasedDecoder import \
    StdTreeDecoder

from graph4nlp.pytorch.modules.utils.tree_utils import to_cuda

from graph4nlp.pytorch.modules.prediction.generation.TreeBasedDecoder import StdTreeDecoder, create_mask
from graph4nlp.pytorch.modules.utils.tree_utils import DataLoaderForGraphEncoder, Tree, Vocab, to_cuda

import warnings

warnings.filterwarnings('ignore')


class Graph2Tree(nn.Module):
    def __init__(self, src_vocab,
                 tgt_vocab,
                 use_copy,
                 enc_hidden_size,
                 dec_hidden_size,
                 dropout_for_decoder,
                 dropout_for_word_embedding,
                 enc_dropout_for_feature,
                 enc_dropout_for_attn,
                 direction_option,
                 input_size,
                 output_size,
                 device,
                 teacher_force_ratio,
                 max_dec_seq_length,
                 max_dec_tree_depth,
                 graph_construction_type,
                 gnn_type,
                 rnn_type):
        super(Graph2Tree, self).__init__()

        self.src_vocab = src_vocab
        self.tgt_vocab = tgt_vocab
        self.device = device
        self.use_copy = use_copy
        self.use_edge_weight = False

        # embedding_style = {'word_emb_type': 'w2v', 'node_edge_emb_strategy': "mean",
        #                    'seq_info_encode_strategy': "bilstm"}

        embedding_style = {'single_token_item': True,
                           'emb_strategy': "w2v_bilstm",
                           'num_rnn_layers': 1
                           }

        if graph_construction_type == "DependencyGraph":
            self.graph_topology = DependencyBasedGraphConstruction(embedding_style=embedding_style,
                                                                   vocab=self.src_vocab,
                                                                   hidden_size=enc_hidden_size, word_dropout=dropout_for_word_embedding, rnn_dropout=dropout_for_word_embedding, device=device,
                                                                   fix_word_emb=False)
        elif graph_construction_type == "ConstituencyGraph":
            self.graph_topology = ConstituencyBasedGraphConstruction(embedding_style=embedding_style,
                                                                     vocab=self.src_vocab,
                                                                     hidden_size=enc_hidden_size, word_dropout=dropout_for_word_embedding, rnn_dropout=dropout_for_word_embedding, device=device,
                                                                     fix_word_emb=False)
        elif graph_construction_type == "DynamicGraph_node_emb":
            self.graph_topology = NodeEmbeddingBasedGraphConstruction(
                self.src_vocab,
                embedding_style,
                sim_metric_type='weighted_cosine',
                num_heads=1,
                top_k_neigh=None,
                epsilon_neigh=0.5,
                smoothness_ratio=0.1,
                connectivity_ratio=0.05,
                sparsity_ratio=0.1,
                input_size=enc_hidden_size,
                hidden_size=enc_hidden_size,
                fix_word_emb=False,
                word_dropout=dropout_for_word_embedding,
                rnn_dropout=dropout_for_word_embedding,
                device=device)
            self.use_edge_weight = True
        elif graph_construction_type == "DynamicGraph_node_emb_refined":
            self.graph_topology = NodeEmbeddingBasedRefinedGraphConstruction(
                self.src_vocab,
                embedding_style,
                0.2,
                sim_metric_type="weighted_cosine",
                num_heads=1,
                top_k_neigh=None,
                epsilon_neigh=0.5,
                smoothness_ratio=0.1,
                connectivity_ratio=0.05,
                sparsity_ratio=0.1,
                input_size=enc_hidden_size,
                hidden_size=enc_hidden_size,
                fix_word_emb=False,
                word_dropout=dropout_for_word_embedding,
                rnn_dropout=dropout_for_word_embedding,
                device=device)
            self.use_edge_weight = True
        else:
            raise NotImplementedError()
            # self.graph_topology = NodeEmbeddingBasedGraphConstruction(word_vocab=self.src_vocab,
            #                                                     embedding_styles=embedding_style,
            #                                                     input_size=enc_hidden_size,
            #                                                     hidden_size=enc_hidden_size,
            #                                                     top_k_neigh=200,
            #                                                     device=device)

        self.word_emb = self.graph_topology.embedding_layer.word_emb_layers[
            'w2v'].word_emb_layer

        if gnn_type == "GAT":
            self.encoder = GAT(1, enc_hidden_size, enc_hidden_size, enc_hidden_size, [1],
                               direction_option=direction_option, feat_drop=enc_dropout_for_feature,
                               attn_drop=enc_dropout_for_attn, activation=F.relu, residual=True)
        elif gnn_type == "GGNN":
            self.encoder = GGNN(1, enc_hidden_size, enc_hidden_size,
                                dropout=enc_dropout_for_feature, use_edge_weight=self.use_edge_weight,
                                direction_option=direction_option)
        elif gnn_type == "SAGE":
            # aggregate type: 'mean','gcn','pool','lstm'
            self.encoder = GraphSAGE(1, enc_hidden_size, enc_hidden_size, enc_hidden_size,
                                     'lstm', direction_option=direction_option, feat_drop=enc_dropout_for_feature,
                                     activation=F.relu, bias=True, use_edge_weight=self.use_edge_weight)
        elif gnn_type == "GCN":
            self.encoder = GCN(1,
                               enc_hidden_size,
                               enc_hidden_size,
                               enc_hidden_size,
                               direction_option=direction_option,
                               norm="both",
                               activation=F.relu,
                               use_edge_weight=self.use_edge_weight)
        else:
            print("Wrong gnn type, please use GAT GGNN or SAGE")
            raise NotImplementedError()
        self.criterion = nn.NLLLoss(size_average=False)

        self.decoder = StdTreeDecoder(attn_type="uniform",
                                      embeddings=self.word_emb,
                                      enc_hidden_size=enc_hidden_size,
                                      dec_emb_size=self.tgt_vocab.embedding_dims,
                                      dec_hidden_size=dec_hidden_size,
                                      output_size=output_size,
                                      device=device,
                                      criterion=self.criterion,
                                      teacher_force_ratio=teacher_force_ratio,
                                      use_sibling=False,
                                      use_attention=True,
                                      use_copy=self.use_copy,
                                      use_coverage=True,
                                      fuse_strategy="average",
                                      num_layers=1,
                                      dropout_for_decoder=dropout_for_decoder,
                                      rnn_type=rnn_type,
                                      max_dec_seq_length=max_dec_seq_length,
                                      max_dec_tree_depth=max_dec_tree_depth,
                                      tgt_vocab=self.tgt_vocab)

    def forward(self, graph_list, tgt_tree_batch, oov_dict=None):
        batch_graph = self.graph_topology(graph_list)
        batch_graph = self.encoder(batch_graph)
        batch_graph.node_features["rnn_emb"] = batch_graph.node_features['node_feat']

        batch_graph_list_decoder_input = from_batch(batch_graph)
        if self.use_copy and "token_id_oov" not in batch_graph.node_features.keys():
            for g, g_ in zip(batch_graph_list_decoder_input, graph_list):
                g.node_features['token_id_oov'] = g_.node_features['token_id_oov']

        loss = self.decoder(g=batch_graph_list_decoder_input,
                            tgt_tree_batch=tgt_tree_batch, oov_dict=oov_dict)
        return loss

    def init(self, init_weight):
        to_cuda(self.encoder, self.device)
        to_cuda(self.decoder, self.device)

        print('--------------------------------------------------------------')
        for name, param in self.named_parameters():
            # print(name, param.size())
            if param.requires_grad:
                if ("word_embedding" in name) or ("word_emb_layer" in name) or ("bert_embedding" in name):
                    pass
                else:
                    if len(param.size()) >= 2:
                        if "rnn" in name:
                            init.orthogonal_(param)
                        else:
                            init.xavier_uniform_(param, gain=1.0)
                    else:
                        init.uniform_(param, -init_weight, init_weight)
        # print('--------------------------------------------------------------')


class Geo:
    def __init__(self, opt=None):
        super(Geo, self).__init__()
        self.opt = opt

        seed = opt.seed
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.backends.cudnn.deterministic = True

        if opt.gpuid == -1:
            self.device = torch.device("cpu")
        else:
            self.device = torch.device("cuda:{}".format(opt.gpuid))
        self.use_copy = True if opt.use_copy == 1 else False
        self.revectorization = True
        self.data_dir = opt.data_dir
        self.checkpoint_dir = opt.checkpoint_dir

        self._build_dataloader()
        self._build_model()
        self._build_optimizer()

    def _build_dataloader(self):
        use_copy = self.use_copy
        use_share_vocab = True

        if self.opt.graph_construction_type == "DependencyGraph":
            dataset = GeoDatasetForTree(root_dir=self.data_dir,
                                         topology_builder=DependencyBasedGraphConstruction,
                                         topology_subdir='DependencyGraph', edge_strategy='as_node',
                                         share_vocab=use_share_vocab, enc_emb_size=self.opt.enc_emb_size,
                                         dec_emb_size=self.opt.tgt_emb_size, device=self.device,
                                         min_freq=self.opt.min_freq)

        elif self.opt.graph_construction_type == "ConstituencyGraph":
            dataset = GeoDatasetForTree(root_dir=self.data_dir,
                                         topology_builder=ConstituencyBasedGraphConstruction,
                                         topology_subdir='ConstituencyGraph', share_vocab=use_share_vocab,
                                         enc_emb_size=self.opt.enc_emb_size, dec_emb_size=self.opt.tgt_emb_size,
                                         device=self.device, min_freq=self.opt.min_freq)

        elif self.opt.graph_construction_type == "DynamicGraph_node_emb":
            dataset = GeoDatasetForTree(root_dir=self.data_dir, seed=self.opt.seed, word_emb_size=self.opt.enc_emb_size,
                                         topology_builder=NodeEmbeddingBasedGraphConstruction,
                                         topology_subdir='DynamicGraph_node_emb', graph_type='dynamic',
                                         dynamic_graph_type='node_emb', share_vocab=use_share_vocab,
                                         enc_emb_size=self.opt.enc_emb_size, dec_emb_size=self.opt.tgt_emb_size,
                                         device=self.device, min_freq=self.opt.min_freq)

        elif self.opt.graph_construction_type == "DynamicGraph_node_emb_refined":
            if self.opt.dynamic_init_graph_type is None or self.opt.dynamic_init_graph_type == 'line':
                dynamic_init_topology_builder = None
            elif self.opt.dynamic_init_graph_type == 'dependency':
                dynamic_init_topology_builder = DependencyBasedGraphConstruction
            elif self.opt.dynamic_init_graph_type == 'constituency':
                dynamic_init_topology_builder = ConstituencyBasedGraphConstruction
            else:
                # dynamic_init_topology_builder
                raise RuntimeError(
                    'Define your own dynamic_init_topology_builder')
            dataset = GeoDatasetForTree(root_dir=self.data_dir, seed=self.opt.seed, word_emb_size=self.opt.enc_emb_size,
                                         topology_builder=NodeEmbeddingBasedRefinedGraphConstruction,
                                         topology_subdir='DynamicGraph_node_emb_refined', graph_type='dynamic',
                                         dynamic_graph_type='node_emb_refined', share_vocab=use_share_vocab,
                                         enc_emb_size=self.opt.enc_emb_size, dec_emb_size=self.opt.tgt_emb_size,
                                         dynamic_init_topology_builder=dynamic_init_topology_builder, device=self.device,
                                         min_freq=self.opt.min_freq)
        else:
            raise NotImplementedError

        self.train_data_loader = DataLoaderForGraphEncoder(
            use_copy=use_copy, use_share_vocab=use_share_vocab, data=dataset.train, dataset=dataset, mode="train", batch_size=self.opt.batch_size, device=self.device)
        print("train sample size:", len(self.train_data_loader.data))
        self.test_data_loader = DataLoaderForGraphEncoder(
            use_copy=use_copy, use_share_vocab=use_share_vocab, data=dataset.test, dataset=dataset, mode="test", batch_size=1, device=self.device)
        print("test sample size:", len(self.test_data_loader.data))

        self.src_vocab = self.train_data_loader.src_vocab
        self.tgt_vocab = self.train_data_loader.tgt_vocab
        # print(self.src_vocab.symbol2idx)
        # print(self.tgt_vocab.symbol2idx)
        if use_share_vocab:
            self.share_vocab = self.train_data_loader.share_vocab
        print(self.share_vocab.symbol2idx)
        print("---Loading data done---\n")

    def _build_model(self):
        '''For encoder-decoder'''
        self.model = Graph2Tree(src_vocab=self.src_vocab,
                                tgt_vocab=self.tgt_vocab,
                                use_copy=self.use_copy,
                                enc_hidden_size=self.opt.enc_hidden_size,
                                dec_hidden_size=self.opt.dec_hidden_size,
                                dropout_for_decoder=self.opt.dropout_for_decoder,
                                dropout_for_word_embedding=self.opt.dropout_for_word_embedding,
                                enc_dropout_for_feature=self.opt.dropout_for_encoder,
                                enc_dropout_for_attn=self.opt.dropout_for_encoder,
                                direction_option=self.opt.direction_option,
                                input_size=self.src_vocab.vocab_size,
                                output_size=self.tgt_vocab.vocab_size,
                                device=self.device,
                                teacher_force_ratio=self.opt.teacher_force_ratio,
                                max_dec_seq_length=self.opt.max_dec_seq_length,
                                max_dec_tree_depth=self.opt.max_dec_tree_depth_for_train,
                                graph_construction_type=self.opt.graph_construction_type,
                                gnn_type=self.opt.gnn_type,
                                rnn_type=self.opt.rnn_type)
        self.model.init(self.opt.init_weight)
        self.model = to_cuda(self.model, self.device)
        print(self.model)

    def _build_optimizer(self):
        optim_state = {"learningRate": self.opt.learning_rate,
                       "weight_decay": self.opt.weight_decay}
        parameters = [p for p in self.model.parameters() if p.requires_grad]
        self.optimizer = optim.Adam(
            parameters, lr=optim_state['learningRate'], weight_decay=optim_state['weight_decay'])
        # self.scheduler = torch.optim.lr_scheduler.StepLR(self.optimizer, step_size = self.opt.max_epochs//3, gamma=0.5)

    def prepare_ext_vocab(self, batch_graphs, src_vocab):
        oov_dict = copy.deepcopy(src_vocab)
        for g in batch_graphs:
            token_matrix = []
            for node_idx in range(g.get_node_num()):
                node_token = g.node_attributes[node_idx]['token']
                if (g.node_attributes[node_idx].get('type') == None or g.node_attributes[node_idx].get('type') == 0) \
                        and oov_dict.get_symbol_idx(node_token) == oov_dict.get_symbol_idx(oov_dict.unk_token):
                    oov_dict.add_symbol(node_token)
                token_matrix.append([oov_dict.get_symbol_idx(node_token)])
            token_matrix = torch.tensor(
                token_matrix, dtype=torch.long).to(self.device)
            g.node_features['token_id_oov'] = token_matrix
        return oov_dict

    def train_epoch(self, epoch):
        loss_to_print = 0
        for i in range(self.train_data_loader.num_batch):
            self.optimizer.zero_grad()
            batch_graph_list, _, batch_tree_list, batch_original_tree_list = self.train_data_loader.random_batch()

            oov_dict = self.prepare_ext_vocab(
                batch_graph_list, self.src_vocab) if self.use_copy else None


            if self.use_copy and self.revectorization:
                batch_tree_list_refined = []
                for item in batch_original_tree_list:
                    tgt_list = oov_dict.get_symbol_idx_for_list(item.strip().split())
                    tgt_tree = Tree.convert_to_tree(tgt_list, 0, len(tgt_list), oov_dict)
                    batch_tree_list_refined.append(tgt_tree)
            # for index in range(len(batch_tree_list_refined)):
            #     print("---------------------------------------")
            #     print(batch_tree_list[index])
            #     print(batch_tree_list_refined[index])
            # loss = self.model(batch_graph_list, batch_tree_list, oov_dict=oov_dict)
            loss = self.model(batch_graph_list, batch_tree_list_refined if self.use_copy else batch_tree_list, oov_dict=oov_dict)
            loss.backward()
            torch.nn.utils.clip_grad_value_(
                self.model.parameters(), self.opt.grad_clip)
            self.optimizer.step()
            loss_to_print += loss
        return loss_to_print/self.train_data_loader.num_batch

    def train(self):
        best_acc = -1

        print("-------------\nStarting training.")
        for epoch in range(1, self.opt.max_epochs+1):
            self.model.train()
            loss_to_print = self.train_epoch(epoch)
            # self.scheduler.step()
            print("epochs = {}, train_loss = {:.3f}".format(epoch, loss_to_print))
            # print(self.scheduler.get_lr())
            if epoch > 20 and epoch % 10 == 0:
                # torch.save(checkpoint, "{}/g2t".format(self.checkpoint_dir) + str(i))
                # pickle.dump(checkpoint, open("{}/g2t".format(self.checkpoint_dir) + str(i), "wb"))
                test_acc = self.eval((self.model))
                if test_acc > best_acc:
                    best_acc = test_acc
        print("Best Acc: {:.3f}\n".format(best_acc))

    def eval(self, model):
        device = model.device

        # max_dec_seq_length = self.opt.max_dec_seq_length
        # max_dec_tree_depth = self.opt.max_dec_tree_depth_for_test
        max_dec_seq_length = 50
        max_dec_tree_depth = 20
        
        use_copy = self.test_data_loader.use_copy
        enc_emb_size = model.src_vocab.embedding_dims
        tgt_emb_size = model.tgt_vocab.embedding_dims

        enc_hidden_size = model.decoder.enc_hidden_size
        dec_hidden_size = model.decoder.hidden_size

        model.eval()

        reference_list = []
        candidate_list = []

        data = self.test_data_loader.data

        for i in range(len(data)):
            x = data[i]

            # get input graph list
            input_graph_list = [x[0]]
            # if use_copy:
            oov_dict = self.prepare_ext_vocab(
                input_graph_list, self.test_data_loader.src_vocab)

            # get indexed tgt sequence
            if self.use_copy and self.revectorization:
                reference = oov_dict.get_symbol_idx_for_list(x[1].split())
                # reference = Tree.convert_to_tree(tmp_list, 0, len(tmp_list), oov_dict)
                eval_vocab = oov_dict

            else:
                reference = model.tgt_vocab.get_symbol_idx_for_list(x[1].split())
                eval_vocab = self.test_data_loader.tgt_vocab

            candidate = model.decoder.translate(use_copy,
                                                enc_hidden_size,
                                                dec_hidden_size,
                                                model,
                                                input_graph_list,
                                                self.test_data_loader.src_vocab,
                                                self.test_data_loader.tgt_vocab,
                                                device,
                                                max_dec_seq_length,
                                                max_dec_tree_depth,
                                                oov_dict=oov_dict)
            
            candidate = [int(c) for c in candidate]
            num_left_paren = sum(
                1 for c in candidate if eval_vocab.idx2symbol[int(c)] == "(")
            num_right_paren = sum(
                1 for c in candidate if eval_vocab.idx2symbol[int(c)] == ")")
            diff = num_left_paren - num_right_paren
            if diff > 0:
                for i in range(diff):
                    candidate.append(
                        self.test_data_loader.tgt_vocab.symbol2idx[")"])
            elif diff < 0:
                candidate = candidate[:diff]
            ref_str = convert_to_string(
                reference, eval_vocab)
            cand_str = convert_to_string(
                candidate, eval_vocab)

            # for c in candidate:
            #     if c >= self.test_data_loader.tgt_vocab.vocab_size:
            #         print("====================")
            #         print(oov_dict.symbol2idx)
            #         print(cand_str)
            #         print(ref_str)
            #         print("====================")
            # if cand_str.strip() != ref_str.strip():
            #     print(cand_str)
            #     print(ref_str)

            reference_list.append(reference)
            candidate_list.append(candidate)
            # print(cand_str)

        test_acc = compute_tree_accuracy(
            candidate_list, reference_list, eval_vocab)
        print("TEST ACCURACY = {:.3f}\n".format(test_acc))
        return test_acc


def convert_to_string(idx_list, form_manager):
    w_list = []
    for i in range(len(idx_list)):
        w_list.append(form_manager.get_idx_symbol(int(idx_list[i])))
    return " ".join(w_list)

def is_all_same(c1, c2, form_manager):
    all_same = True
    if len(c1) == len(c2):
        all_same = True
        for j in range(len(c1)):
            if c1[j] != c2[j]:
                all_same = False
                break
        if all_same:
            return True
    if len(c1) != len(c2) or all_same == False:
        n1 = Tree.deduplicate_tree(Tree.norm_tree(c1, form_manager), form_manager)
        n2 = Tree.deduplicate_tree(Tree.norm_tree(c2, form_manager), form_manager)
        # if np.array_equal(np.array(c1), np.array(n1)) == False:
        #     print(form_manager.get_idx_symbol_for_list(c1))
        #     print(form_manager.get_idx_symbol_for_list(n1))
        #     print("=================")
        if len(n1) == len(n2):
            all_same = True
            for j in range(len(n1)):
                if n1[j] != n2[j]:
                    all_same = False
                    break
        else:
            return False
        return all_same
    raise NotImplementedError("you should not arrive here!")

def compute_accuracy(candidate_list, reference_list, form_manager):
    if len(candidate_list) != len(reference_list):
        print("candidate list has length {}, reference list has length {}\n".format(
            len(candidate_list), len(reference_list)))

    len_min = min(len(candidate_list), len(reference_list))
    c = 0
    for i in range(len_min):
        if is_all_same(candidate_list[i], reference_list[i], form_manager):
            c = c+1
        else:
            pass

    return c/float(len_min)


def compute_tree_accuracy(candidate_list_, reference_list_, form_manager):
    candidate_list = []
    for i in range(len(candidate_list_)):
        candidate_list.append(candidate_list_[i])
    reference_list = []
    for i in range(len(reference_list_)):
        reference_list.append(reference_list_[i])
    return compute_accuracy(candidate_list, reference_list, form_manager)


if __name__ == "__main__":
    start = time.time()
    main_arg_parser = argparse.ArgumentParser(description="parser")

    main_arg_parser.add_argument(
        '-gpuid', type=int, default=0, help='which gpu to use. -1 = use CPU')
    main_arg_parser.add_argument(
        '-seed', type=int, default=123, help='torch manual random number generator seed')
    main_arg_parser.add_argument(
        '-use_copy', type=int, default=0, help='whether use copy mechanism')

    main_arg_parser.add_argument('-data_dir', type=str,
                                 default='/home/lishucheng/Graph4AI/graph4nlp/examples/pytorch/semantic_parsing/graph2tree/data/geo', help='data path')
    main_arg_parser.add_argument('-checkpoint_dir', type=str,
                                 default='/home/lishucheng/Graph4AI/graph4nlp/examples/pytorch/semantic_parsing/graph2tree/checkpoint_dir_geo', help='output directory where checkpoints get written')

    main_arg_parser.add_argument('-gnn_type', type=str, default="SAGE")
    main_arg_parser.add_argument('-rnn_type', type=str, default="lstm")

    main_arg_parser.add_argument('-enc_emb_size', type=int, default=300)
    main_arg_parser.add_argument('-tgt_emb_size', type=int, default=300)

    main_arg_parser.add_argument('-enc_hidden_size', type=int, default=300)
    main_arg_parser.add_argument('-dec_hidden_size', type=int, default=300)

    # DynamicGraph_node_emb_refined, DynamicGraph_node_emb
    main_arg_parser.add_argument(
        '-graph_construction_type', type=str, default="ConstituencyGraph")

    # "None, line, dependency, constituency"
    main_arg_parser.add_argument(
        '-dynamic_init_graph_type', type=str, default="constituency")

    main_arg_parser.add_argument('-batch_size', type=int, default=20)

    main_arg_parser.add_argument(
        '-dropout_for_word_embedding', type=float, default=0.1)

    main_arg_parser.add_argument(
        '-dropout_for_encoder', type=float, default=0)

    main_arg_parser.add_argument(
        '-dropout_for_decoder', type=float, default=0.1)

    main_arg_parser.add_argument(
        '-direction_option', type=str, default="undirected")

    main_arg_parser.add_argument('-max_dec_seq_length', type=int, default=100)
    main_arg_parser.add_argument(
        '-max_dec_tree_depth_for_train', type=int, default=25)
    main_arg_parser.add_argument(
        '-max_dec_tree_depth_for_test', type=int, default=25)

    main_arg_parser.add_argument(
        '-teacher_force_ratio', type=float, default=1.0)

    main_arg_parser.add_argument(
        '-init_weight', type=float, default=0.08, help='initailization weight')
    main_arg_parser.add_argument(
        '-learning_rate', type=float, default=1e-3, help='learning rate')
    main_arg_parser.add_argument('-weight_decay', type=float, default=0)

    main_arg_parser.add_argument('-max_epochs', type=int, default=150,
                                 help='number of full passes through the training data')
    main_arg_parser.add_argument('-min_freq', type=int, default=1,
                                 help='minimum frequency for vocabulary')
    main_arg_parser.add_argument(
        '-grad_clip', type=int, default=5, help='clip gradients at this value')

    args = main_arg_parser.parse_args()

    runner = Geo(opt=args)
    max_score = runner.train()

    end = time.time()
    print("total time: {} minutes\n".format((end - start)/60))
